/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#include "AStar.h"
#include "Game.h"
#include "Map.h"
#include "Tile.h"
#include "TileLocation.h"

AStar::AStar() : 
    numTilesX(g_game->getMap()->getMapWidth()),
    numTilesY(g_game->getMap()->getMapLength()) {
}

AStar::~AStar() {
    clear();
}

void AStar::clear() {
    for( openIt = openList.begin(); openIt != openList.end(); ++openIt ) {
        delete openIt->second;
    }
    for( closedIt = closedList.begin(); closedIt != closedList.end(); ++closedIt ) {
        delete closedIt->second;
    }

    openList.clear();
    closedList.clear();
    sortedList.clear();
    path.clear();
}

bool AStar::findpath( const Point& source, const Point& dest ) {
    clear();

    sourceX = (int)source.x;
    sourceY = (int)source.y;
    destX = (int)dest.x;
    destY = (int)dest.y;

    /*TRACE2( "source: %d,%d", sourceX, sourceY );
    TRACE2( "dest: %d,%d", destX, destY );*/

    int h = getManhattanDistance( sourceX, sourceY );
    AStarTile* start = new AStarTile( sourceX, sourceY, 0, h, NULL );
    sortedList.insert( AStarTileKey(start) );
    openList[start->getPoint()] = start;
    Map* map = g_game->getMap();

    while( !openList.empty() ) {
        AStarTileKey key = *sortedList.begin();
        AStarTile* best = key.getTile();
        sortedList.erase( sortedList.begin() );
        openList.erase( best->getPoint() );
        TileLocation parentLoc( best->getX(), best->getY() );

        // Check to see if a path has been found
        if( (best->getX() == destX) && (best->getY() == destY) ) {
            AStarTile* tile = best;

            // Create the path
            while( (tile->getX() != sourceX) || (tile->getY() != sourceY) ) {
                // Add 0.5 to get the bomb to move to the center of the tile
                path.push_front( Point(tile->getX() + 0.5f, tile->getY() + 0.5f) );
                tile = tile->getParent();
            }

            /*std::list<Point>::iterator it;
            for( it = path.begin(); it != path.end(); it++ ) {
            TRACE2( "path: %f,%f", it->x, it->y );
            }*/

            return true;
        }

        // Process each neighboring tile
        for( int i = 0; i < 4; i++ ) {
            TileLocation testLoc(parentLoc);  

            switch(i) {
            case 0:
                testLoc.northeast();
                break;
            case 1:
                testLoc.northwest();
                break;
            case 2:
                testLoc.southeast();
                break;
            case 3:
                testLoc.southwest();
                break;
            }

            if( !map->isLocationOutside(testLoc) ) {
                processTile( best, testLoc.getX(), testLoc.getY() );
            }
        }

        closedList[best->getPoint()] = best;
    }

    // No path found
    return false;
}

int AStar::getManhattanDistance( int x, int y ) {
    return abs( destX - x ) + abs( destY - y );
}

void AStar::processTile( AStarTile* parent, int x, int y ) {
    int g = parent->getG() + 1;

    Point parentPt = parent->getPoint();
    Point pt = Point((float)x , (float)y);
    const Tile* mapTileP = g_game->getMap()->getTileAt( parentPt );
    const Tile* mapTile = g_game->getMap()->getTileAt( pt );
    // If the tile is more than 1 higher than its parent,
    // mark it as impassable by giving it a high fitness
    if( mapTile->height > (mapTileP->height + 1) ) {
        g += numTilesX * numTilesY;
    }

    int h = getManhattanDistance( x, y );
    AStarTile* tile = new AStarTile( x, y, g, h, parent );
    AStarTileKey key( tile );

    openIt = openList.find( tile->getPoint() );
    closedIt = closedList.find( tile->getPoint() );

    if( openIt != openList.end() ) {
        AStarTile* found = openIt->second;

        if( tile->getF() < found->getF() ) {
            openList[tile->getPoint()] = tile;
            sortedIt = std::find( sortedList.begin(), sortedList.end(), AStarTileKey(found) );
            sortedList.erase( sortedIt );
            sortedList.insert( AStarTileKey(tile) );
            delete found;
        }
		else {
			delete tile;
		}
    }
    else if( closedIt != closedList.end() ) {
        AStarTile* found = closedIt->second;

        if( tile->getF() < found->getF() ) {
            openList.erase( found->getPoint() );
            openList[tile->getPoint()] = tile;
            sortedList.insert( AStarTileKey(tile) );
            delete found;
        }
		else {
			delete tile;
		}
    }
    else {
        openList[tile->getPoint()] = tile;
        sortedList.insert( AStarTileKey(tile) );
    }
}

Point AStar::getNextPoint() {
    Point pt( path.front() );
    path.pop_front();
    return pt;
}

bool AStar::hasMorePoints() {
    return !path.empty();
}
