/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#ifndef _ASTARTILE_H_
#define _ASTARTILE_H_

#include "IsoBomb.h"

class AStarTile {
    
/**
 * Represents a tile or node in the A* algorithm.
 */
public:

    /**
     * Constructor
     */
    AStarTile();

    /**
     * Constructor
     *
     * @param initX                 the x coordinate
     * @param initY                 the y coordinate
     * @param initG                 the g value in the A* algorithm
     * @param initH                 the h value in the A* algorithm
     * @param initParent            the parent tile
     */
    AStarTile( int initX, int initY, int initG, int initH, AStarTile* initParent );

    /**
     * Destructor
     */
    ~AStarTile();

    /**
     * Initializes the tile.
     *
     * @param initX                 the x coordinate
     * @param initY                 the y coordinate
     * @param initG                 the g value in the A* algorithm
     * @param initH                 the h value in the A* algorithm
     * @param initParent            the parent tile
     */
    void init( int initX, int initY, int initG, int initH, AStarTile* initParent );

    /**
     * Returns the f value of the A* algorithm (fitness).
     *
     * @returns                     f value of the A* algorithm (fitness)
     */
    int getF() const;

    /**
     * Returns the g value of the A* algorithm (goal).
     *
     * @returns                     g value of the A* algorithm (goal)
     */
    int getG() const;

    /**
     * Returns the h value of the A* algorithm (heuristic).
     *
     * @returns                     h value of the A* algorithm (heuristic)
     */
    int getH() const;

    /**
     * Returns the x coordinate of this tile.
     *
     * @returns                     the x coordinate of this tile.
     */
    int getX() const;

    /**
     * Returns the y coordinate of this tile.
     *
     * @returns                     the y coordinate of this tile.
     */
    int getY() const;

    /**
     * Returns a pointer to this tile's parent tile.
     *
     * @returns                     a pointer to this tile's parent tile
     */
    AStarTile* getParent() const;


    Point getPoint() const;

private:

    int g;

    int h;

    int f;

    Point pt;

    AStarTile* parent;

};

/**
 * Used to store AStarTile pointers in a collection.  Implements
 * the < operator so that sorting is done by fitness.  Implements
 * the == operator so that finding is done by coordinates.
 */
class AStarTileKey {

public:

    /**
     * Constructor
     *
     * @param theTile               pointer to the AStarTile to hold
     */
    AStarTileKey( AStarTile* theTile ) : tile(theTile) {
        
    }

    /**
     * Destructor
     */
    ~AStarTileKey() {

    }

    /**
     * Returns true if this key's tile's fitness is less than
     * the other key's tile's fitness.  If it is the same fitness, 
     * then compare by the x and y values so that set will allow
     * keys of the same fitness to exist in the set.
     */
    bool operator<( const AStarTileKey rhs ) const {
        AStarTile* tile2 = rhs.getTile();

        if( tile->getF() == tile2->getF() ) {
            if( tile->getX() != tile2->getX() ) {
                return tile->getX() < tile2->getX();
            }
            else {
                return tile->getY() < tile2->getY();
            }
        }
        else {
            return tile->getF() < tile2->getF();
        }
    }

    /**
     * Return true if the key's tiles are at the same coordinates.
     */
    bool operator==( const AStarTileKey rhs ) const {
        return (tile->getX() == rhs.getTile()->getX()) &&
               (tile->getY() == rhs.getTile()->getY());
    }
    
    /**
     * Returns the tile pointer.
     *
     * @returns                     the tile pointer
     */
    AStarTile* getTile() const {
        return tile;
    }

private:

    AStarTile* tile;
    
};

#endif