/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#include "IsoBomb.h"
#include "Character.h"
#include "Game.h"
#include "Bomb.h"
#include "HomingBomb.h"
#include "ThrownBomb.h"
#include "ConfigFile.h"
#include "TextureManager.h"
#include "Explosion.h"
#include "Powerup.h"
#include "Map.h"
#include "EndGame.h"
#include "ConfigFile.h"
#include "Virus.h"

#include "Particle.h"
#include "Vector3D.h"
#include "Point3D.h"

float Character::invincibleTime = 3.0f;

/**
 * The minimum throwing power.
 */
const float MIN_POWER = 0.30f;

void Character::initVars( ConfigFile& f ) {
    f.getVar( "Character", "invincibleTime", invincibleTime, 3.0f );
}


Character::Character() : power(MIN_POWER), numBombs(0), enemy(NULL), invincibleTimer(0.0f), 
 dead(false), spinSpeed(0.4f), virus(NULL),
 spinDirection(0.0f), totalTimeSinceDeath(0.0f), numWins(0) {

    resetProps();

    anim.setOffsets( 32, 96 );
	//anim.setUsingAlpha( false );
    anim.setAnimationFrames(g_am->getCharacterAnimationFrames(direction));
    anim.setAnimRuntime( 1 / props.speed );
}

Character::Character( const Rect &initPos, int initDir, int wins ) : 
 Entity(initPos, initDir), power(MIN_POWER), numBombs(0), numWins(wins),
 enemy(NULL), invincibleTimer(0.0f), lastPos(initPos), dead( false ),
 spinDirection(0.0f), totalTimeSinceDeath(0.0f), virus(NULL) {

    resetProps();

    anim.setOffsets( 32, 96 );
    anim.setAnimationFrames(g_am->getCharacterAnimationFrames(direction));
    anim.setAnimRuntime( 1 / props.speed );


    HRESULT r = D3DXCreateSprite( g_dx->getDeviceHandle(), &hudSprite );
    if( FAILED(r) ) {
        TRACE( "Failed to create hud sprite" );
    }

    cachedHeight = (float)g_game->getMap()->getRectHeight( pos );

    // Register the listener, so switching modes will work correctly
    g_dx->registerListener( *this );
}

Character::~Character() {
    // Remove the listener
    g_dx->removeListener( *this );
}

void Character::resetProps() {
    // Default properties
    assert( g_cfg != NULL );
    g_cfg->getVar( "Character", "blastSize",       props.blastSize,       2.0f );
    g_cfg->getVar( "Character", "detonationDelay", props.detonationDelay, 5.0f );
    g_cfg->getVar( "Character", "hp",              props.hp,              1 );
    g_cfg->getVar( "Character", "maxBombs",        props.maxBombs,        1 );
    g_cfg->getVar( "Character", "maxPower",        props.maxPower,        1.0f );
    g_cfg->getVar( "Character", "speed",           props.speed,           3.0f );
    g_cfg->getVar( "Character", "powerRate",       props.powerRate,       0.75f );
    g_cfg->getVar( "Character", "numHoming",       props.numHoming,       50 );
    g_cfg->getVar( "Character", "homingSpeed",     props.homingSpeed,     3.0f );
}

bool Character::affectMe( const CharacterProps& modprops ) {
    bool overLimit = false;

    props.blastSize += modprops.blastSize;
    props.detonationDelay += modprops.detonationDelay;
    props.hp += modprops.hp;
    props.maxBombs += modprops.maxBombs;
    props.maxPower += modprops.maxPower;
    props.speed += modprops.speed;
    props.powerRate += modprops.powerRate;
    props.numHoming += modprops.numHoming;
    props.homingSpeed += modprops.homingSpeed;

    // Normalize the props
    if( props.hp > MAX_HP ) {
        props.hp -= modprops.hp;
        overLimit = true;
    }
    if( props.maxPower > MAX_POWER ) {
        props.maxPower -= modprops.maxPower;
        overLimit = true;
    }
    if( props.speed > MAX_SPEED ) {
        props.speed -= modprops.speed;
        overLimit = true;
    }
    if( props.numHoming > MAX_HOMING ) {
        props.numHoming -= modprops.numHoming;
        overLimit = true;
    }
    if( props.maxBombs > MAX_BOMBS ) {
        props.maxBombs -= modprops.maxBombs;
        overLimit = true;
    }

    anim.setAnimRuntime( 1 / props.speed );

    return !overLimit;
}

void Character::chargeThrow( float dt ) {
    if( numBombs < props.maxBombs ) {
        power += (props.powerRate * dt);
        if( power > props.maxPower ) {
            power = props.maxPower;
        }
    }
}

void Character::throwBomb() {
    ThrownBomb* bomb = new ThrownBomb( pos, direction, props.blastSize, 
                                       props.detonationDelay, power, this );
    g_game->addEntity( bomb );

    power = MIN_POWER;
    numBombs++;
}

void Character::homingBomb() {
    assert( enemy != NULL );

    if( props.numHoming != 0 ) {
        Rect bombPos( pos );
        bombPos.setCenter( Point((float)floor(pos.x) + 0.5f, (float)floor(pos.y) + 0.5f) );
        HomingBomb* bomb = new HomingBomb( bombPos, direction, (int)props.blastSize,
            props.detonationDelay, props.homingSpeed, enemy );
        g_game->addEntity( bomb );
        
        props.numHoming--;
    }
}

void Character::move( int dir, float dt ) {
    Rect newPos( pos );
    direction = dir;

    switch( direction ) {
    case DIRECTION_DOWNRIGHT:
        newPos.x += (dt * props.speed);
        break;
    case DIRECTION_DOWNLEFT:
        newPos.y += (dt * props.speed);
        break;
    case DIRECTION_UPRIGHT:
        newPos.y -= (dt * props.speed);
        break;
    case DIRECTION_UPLEFT:
        newPos.x -= (dt * props.speed);
        break;
    default:
        assert( false );
    }

    // Update the position only if this move won't move the chracter off the map
    const Map& m= *(g_game->getMap());
    if( !m.isLocationOutside( newPos ) ) {
        //Update the position only if the player can climb it (or doesn't have to climb)
        int oldHeight = (int)cachedHeight;
        int newHeight = m.getRectHeight( newPos );
        bool isShrunk = false;
        if( virus != NULL ) {
            isShrunk = (virus->getVirusType() == SHRINK_VIRUS);
        }
        
        // Can't climb stairs if shrunk
        if( isShrunk && (newHeight - oldHeight) > 0 ) {
            stopAnimation();
        }
        else if( newHeight - oldHeight <= 1 ) {
            // Remember the last position
            lastPos = pos;
            pos = newPos;
            cachedHeight = (float)g_game->getMap()->getRectHeight( pos );
            anim.start();
        } 
        else {
            stopAnimation();
        }
    }

    anim.setAnimationFrames(g_am->getCharacterAnimationFrames(direction));
}

void Character::noMove( float dt ) {
    //Since we didn't move, our lastPos is this pos.
    lastPos = pos;
}

void Character::draw( const Point& anchor ) {
    // Do the "blinking" effect when invincible
    if( invincibleTimer > 0 && !dead ) {
        int val = (int)(invincibleTimer / 0.15f);
        if( (val % 2) == 0 ) {
            anim.setAlphaValue( 150 );
        }
        else {
            anim.setAlphaValue( 255 );
        }
    }
    else if( invincibleTimer <= 0 && !dead ) {
        anim.setAlphaValue( 255 );
    }

    // Do the blinking red effect when there is a virus
    if( (virus != NULL) && !dead ) {
        anim.setColor( virus->getRedValue(), 50, 50 );
    }
    else {
        anim.setColor( 255, 255, 255 );
    }

    anim.draw( anchor );
}

bool Character::update( float dt ) {
    
    anim.update( dt );

    // Update invincible timer if invincible
    if( invincibleTimer > 0 ) {
        invincibleTimer -= dt;
    }

    // Apply virus's effects if there is one
    if( virus != NULL ) {
        if( virus->update(dt) ) {
            delete virus;
            virus = NULL;
        }
    }

    // The character has died
    if ( (props.hp <= 0) && !dead ) {
        g_snd->playDeathSound();
        g_game->addLogic( new EndGame() );
        dead = true;
        anim.setAlphaValue( 255 );
    }

    // Begin death animation
    if( dead ) {
        // Set color back to normal in case it was changed
        anim.setColor( 255, 255, 255 );

        //Particle spawn because particles are cool!
        for ( int i = 0; i < 3; ++i ) {
            Particle* p = new Particle(
                Point3D( pos.x, pos.y, getHeight() + getrndf( 4.0f, 2.0f ) + (float)(spinDirection / 24) ),
                Vector3D( getrndf( 10.0f ), getrndf( 10.0f ), getrndf( 15.0f ) ),
                10.0f );
            p->setColor( D3DCOLOR_ARGB( 255, getrnd( 255, 192 ), getrnd( 64 ), 0 ) );
            p->setBounceFactor( 0.0f );
            p->setSize( 2.0f );
            g_game->addEffect( p );
        }

        spinSpeed -= dt;
        totalTimeSinceDeath += dt;
        if( spinSpeed < 0 && totalTimeSinceDeath < 4.5f) {
            spinSpeed = 0.4f * ( 1.0f - ( (float)spinDirection * 0.06f ) );

            if( (int)spinDirection % 4 == 0 ) 
                anim.setAnimationFrames( g_am->getCharacterAnimationFrames(DIRECTION_UPLEFT));
            else if( (int)spinDirection % 4 == 1 ) 
                anim.setAnimationFrames( g_am->getCharacterAnimationFrames(DIRECTION_UPRIGHT));			
            else if( (int)spinDirection % 4 == 2 ) 
                anim.setAnimationFrames( g_am->getCharacterAnimationFrames(DIRECTION_DOWNRIGHT));			
            else if( (int)spinDirection % 4 == 3 ) 
                anim.setAnimationFrames( g_am->getCharacterAnimationFrames(DIRECTION_DOWNLEFT));

            spinDirection += dt * 100.0f;
            anim.setOffsets( 32, 96 + (int)spinDirection );
        }
        if( totalTimeSinceDeath >= 4.5 && totalTimeSinceDeath < 25.0) {
            g_snd->playExplosion();
            anim.stop();
            anim.setOffsets( 256, 256 + (int)spinDirection );
            anim.setAlphaValue( 255 ); 
            anim.setUsingAlpha( true );
            anim.setScaleFactor( 8.0f );
            anim.setAnimationFrames( g_am->getExplosionAnimationFrames() );
            anim.setAnimRuntime( 0.5f );
            anim.start();

            totalTimeSinceDeath = 25.0;
        } 
    }

    //The players can't be deleted.  Too many things point to them.
    return false;
}

void Character::stopAnimation() {
    anim.stop();
}

bool Character::isCharging() const {
    return power > MIN_POWER;
}

void Character::setEnemy( const Character* c ) {
    enemy = c;
}

void Character::decBombCounter() {
    assert( numBombs != 0 );
    numBombs--;
}

void Character::decHitPoints() {
    // Make player invincible after a hit so that a single
    // blast can't hurt player multiple times.
    if( invincibleTimer <= 0 ) {
        props.hp -= 1;
        if ( props.hp > 0 )
            g_snd->playHurtSound();
        invincibleTimer = invincibleTime;

        //Particle spawn because particles are cool!
        for ( int c = 0; c < 25; ++c ) {
            Particle* p = new Particle(
                Point3D( pos.x, pos.y, getHeight() + getrndf( 4.0f, 1.0f ) ),
                Vector3D( getrndf( 0.5f, -0.5f ), getrndf( 0.5f, -0.5f ), getrndf( 3.0f ) ),
                getrndf( 1.5f, 1.0f ) );
            p->setColor( D3DCOLOR_ARGB( 255, getrnd( 255, 192 ), getrnd( 64 ), 0 ) );
            p->setBounceFactor( 0.0f );
            p->setSize( 3.0f );
            g_game->addEffect( p );
        }
    }
}

void Character::drawHUD( const Point& anchor ) {
    D3DXVECTOR2 trans;

    // Filled powerbar rect
    RECT rect;
    rect.top = 0;
    rect.left = 0;
    rect.bottom = 32;
    rect.right = (LONG)(((power - MIN_POWER) / (props.maxPower - MIN_POWER)) * 128.0f);

    hudSprite->Begin();

    // Draw the wins counter
    trans.x = anchor.x + 335;
    trans.y = 5;
    int tens = numWins / 10;
    if( tens != 0 ) {
        hudSprite->Draw( g_hudTm->getTexture(tens), // Texture to draw
            0,	    // Rectangle of the part of the texture to draw
            0,		// Scaling amount
            0, 0,	// Rotation center and rotation amount
            &trans, // Translate amounts
            D3DCOLOR_RGBA(255, 255, 255, 180) ); // Color
    }
    trans.x = anchor.x + 363;
    int ones = numWins % 10;
    hudSprite->Draw( g_hudTm->getTexture(ones), // Texture to draw
        0,	    // Rectangle of the part of the texture to draw
        0,		// Scaling amount
        0, 0,	// Rotation center and rotation amount
        &trans, // Translate amounts
        D3DCOLOR_RGBA(255, 255, 255, 180) ); // Color


    // Draw the hp markers
    trans.y = 565;
    for( int i = 0; i < props.hp; i++ ) {
        trans.x = anchor.x + 5 + (i * 30);
        hudSprite->Draw( g_hudTm->getTexture(15), // Texture to draw
            0,	// Rectangle of the part of the texture to draw
            0,		// Scaling amount
            0, 0,	// Rotation center and rotation amount
            &trans, // Translate amounts
            D3DCOLOR_RGBA(255, 255, 255, 180) ); // Color
    }

    trans.x = anchor.x + 158;
    // Draw the power bar outline
    hudSprite->Draw( g_hudTm->getTexture(12), // Texture to draw
        0,	// Rectangle of the part of the texture to draw
        0,		// Scaling amount
        0, 0,	// Rotation center and rotation amount
        &trans, // Translate amounts
        D3DCOLOR_RGBA(255, 255, 255, 180) ); // Color
    // Draw filled power bar
    hudSprite->Draw( g_hudTm->getTexture(13), // Texture to draw
        &rect,	// Rectangle of the part of the texture to draw
        0,		// Scaling amount
        0, 0,	// Rotation center and rotation amount
        &trans, // Translate amounts
        D3DCOLOR_RGBA(255, 255, 255, 180) ); // Color

    // Draw the bomb icon
    trans.x = anchor.x + 285;
    hudSprite->Draw( g_hudTm->getTexture(16), // Texture to draw
        0,	    // Rectangle of the part of the texture to draw
        0,		// Scaling amount
        0, 0,	// Rotation center and rotation amount
        &trans, // Translate amounts
        D3DCOLOR_RGBA(255, 255, 255, 180) ); // Color

    // Draw the number of available bombs
    // Number of bombs can be negative when a no bomb virus is
    // picked up while bombs are laid.
    int availBombs = props.maxBombs - numBombs;
    if( availBombs < 0 ) {
        availBombs = 0;
    }
    trans.x = anchor.x + 310;
    hudSprite->Draw( g_hudTm->getTexture(availBombs), // Texture to draw
        0,	    // Rectangle of the part of the texture to draw
        0,		// Scaling amount
        0, 0,	// Rotation center and rotation amount
        &trans, // Translate amounts
        D3DCOLOR_RGBA(255, 255, 255, 180) ); // Color	


    // Draw the homing bomb icon
    trans.x = anchor.x + 337;
    hudSprite->Draw( g_hudTm->getTexture(11), // Texture to draw
        0,	    // Rectangle of the part of the texture to draw
        0,		// Scaling amount
        0, 0,	// Rotation center and rotation amount
        &trans, // Translate amounts
        D3DCOLOR_RGBA(255, 255, 255, 180) ); // Color

    // Draw the homing bomb count
    trans.x = anchor.x + 366;
    hudSprite->Draw( g_hudTm->getTexture(props.numHoming), // Texture to draw
        0,	    // Rectangle of the part of the texture to draw
        0,		// Scaling amount
        0, 0,	// Rotation center and rotation amount
        &trans, // Translate amounts
        D3DCOLOR_RGBA(255, 255, 255, 180) ); // Color	

    hudSprite->End();
}

void Character::OnLostDevice() {
    hudSprite->OnLostDevice();
}

void Character::OnResetDevice() {
    hudSprite->OnResetDevice();
}

float Character::getHeight() const {
    return cachedHeight;
}

void Character::moveBack() {
    pos = lastPos;
}

void Character::doCollision( Entity* otherEntity ) {
    Powerup* p = dynamic_cast<Powerup*>(otherEntity);
    Character* c = dynamic_cast<Character*>(otherEntity);
    if( p != NULL ) {
        p->affectCharacter( this );
    }
    else if( c != NULL ) {
        moveBack();
        // Inflict character with virus if collision
        if( (virus != NULL) && (c->virus == NULL) ) {
            c->affectMe( virus->getCloneFor(c) );
        }
    }
    else if( dynamic_cast<Bomb*>(otherEntity) != NULL ) {
        moveBack();
    }
    else if( dynamic_cast<Explosion*>(otherEntity) != NULL ) {
        this->decHitPoints();
    }
}

bool Character::isDead() const {
    return dead;
}

CharacterProps Character::getProps() const {
    return props;
}

void Character::affectMe( Virus* v ) {
    virus = v;
}

Virus* Character::getVirus() const {
    return virus;
}

void Character::setShrink( bool shrink ) {
    if( shrink ) {
        anim.setScaleFactor( 0.5f );
        anim.setOffsets( 16, 49 );
        pos.w = 0.2f;
        pos.h = 0.2f;
    }
    else {
        anim.setScaleFactor( 1.0f );
        anim.setOffsets( 32, 96 );
        pos.w = 0.35f;
        pos.h = 0.35f;
    }
}

bool Character::checkCollision( const Entity* otherEntity ) {
    if( pos.isCollision( otherEntity->getPosition() ) ) {
        float diff = fabsf( this->getHeight() - otherEntity->getHeight() );
        bool isShrunk = false;
        if( virus != NULL ) {
            isShrunk = (virus->getVirusType() == SHRINK_VIRUS);
        }

        // If character is shrunk, only collide if within 1 height
        if( isShrunk ) {
            return (diff <= 1.0f);
        }
        else {
            return (diff <= 3.0f);
        }
    } 
    else {
        return false;
    }
}
