/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#include "IsoBomb.h"
#include "Entity.h"
#include "Controller.h"
#include "Map.h"
#include "Character.h"
#include "GameLogic.h"
#include "Game.h"
#include "HomingBomb.h"
#include "ConfigFile.h"
#include "Effect.h"

bool deviceReset = false;

Game::Game() : map(NULL), keyPressed(-1), currMap(-1), resetMapFlag(false),
                numWins1(0), numWins2(0) {

    // Player 1 control settings
    keys1.bomb = DIK_LSHIFT;
    keys1.homing = DIK_TAB;
    keys1.down = DIK_S;
    keys1.left = DIK_A;
    keys1.right = DIK_D;
    keys1.up = DIK_W;

    // Player 2 control settings
    keys2.bomb = DIK_RETURN;
    keys2.homing = DIK_RCONTROL;
    keys2.down = DIK_DOWN;
    keys2.left = DIK_LEFT;
    keys2.right = DIK_RIGHT;
    keys2.up = DIK_UP;
}

Game::~Game() {
    clear();
}

void Game::clear() {
    // Delete all logics
    for( logicIt = logics.begin(); logicIt != logics.end(); ++logicIt ) {
        delete *logicIt;
    }
    // Delete all effects
    for( effectIt = effects.begin(); effectIt != effects.end(); ++effectIt ) {
        delete *effectIt;
    }
    // Delete all entities
    for( entityIt = entities.begin(); entityIt != entities.end(); ++entityIt ) {
        delete *entityIt;
    }
    // Delete all controllers
    for( ctrlIt = ctrls.begin(); ctrlIt != ctrls.end(); ++ctrlIt ) {
        delete *ctrlIt;
    }

    delete map;
    map = NULL;

    effects.clear();
    logics.clear();
    entities.clear();
    ctrls.clear();
}

bool Game::startNewGame( int mapNum ) {
    // Load the specified map
    char mapFile[256];
    sprintf( mapFile, "maps/map%d.map", mapNum );
    Map* newMap = new Map();
    
    if( !newMap->load(mapFile) ) {
        // If the map cannot be loaded delete the map object and 
        // do not continue
        delete newMap;
        return true;
    }
    else {
        // Clean up the existing game
        clear();

        // Set the new map
        map = newMap;
        currMap = mapNum;

        // Get the character starting points
        Point pt1 = map->getStartingPoint();
        Point pt2 = map->getStartingPoint();
        Rect pos1( 0.0f, 0.0f, 0.35f, 0.35f );
        Rect pos2( 0.0f, 0.0f, 0.35f, 0.35f );
        pos1.setCenter( pt1 );
        pos2.setCenter( pt2 );

        // Create the characters
        player1 = new Character( pos1, DIRECTION_DOWNLEFT, numWins1 );
        player2 = new Character( pos2, DIRECTION_DOWNLEFT, numWins2 );
        player1->setEnemy( player2 );
        player2->setEnemy( player1 );

        // Create the controllers
        KeyboardController* ctrl1 = new KeyboardController( player1, keys1 );
        KeyboardController* ctrl2 = new KeyboardController( player2, keys2 );

        // Add the characters and controllers to the game
        addEntity( player1 );
        addEntity( player2 );
        addController( ctrl1 );
        addController( ctrl2 );

        // Setup the cameras
        map->setCameraTarget( player1, 0 );
        map->setCameraTarget( player2, 1 );

        // Add the powerups in the map to the game.
        // Must be done after loading the map since 
        // everything is removed from the game right
        // after the map is loaded.
        map->addPowerups();
        return false;
    }
}

void Game::resetMap() {
    resetMapFlag = true;

    if( !(player1->isDead() && player2->isDead()) ) {
        if( player1->isDead() ) {
            numWins2++;
            if( numWins2 == 100 ) {
                numWins2 = 0;
            }
        }
        else {
            numWins1++;
            if( numWins1 == 100 ) {
                numWins1 = 0;
            }
        }
    }
}

void Game::update( float dt ) {
    assert( map != NULL );

    // Reset the map if it was requested.
    if ( resetMapFlag ) {
        bool failed = startNewGame( currMap );
        assert( !failed );
        resetMapFlag = false;
    }

    // Update all GameLogics
    logicIt = logics.begin();
    while ( logicIt != logics.end() ) {
        if ( (*logicIt)->update( dt ) ) {
            delete (*logicIt);
            logicIt = logics.erase( logicIt );
        } else {
            ++logicIt;
        }
    }

    // Update all Effects
    effectIt = effects.begin();
    while ( effectIt != effects.end() ) {
        if ( (*effectIt)->update( dt ) ) {
            delete (*effectIt);
            effectIt = effects.erase( effectIt );
        } else {
            ++effectIt;
        }
    }

    // Update all entities
    entityIt = entities.begin();
    while( entityIt != entities.end() ) {

        // If this entity has died, remove it		
        if( (*entityIt)->update( dt ) ) {
            delete *entityIt;
            entityIt = entities.erase( entityIt );
        }
        else {
            ++entityIt;
        }
    }

    std::list<Entity *>::iterator entityIt2;
    entityIt = entities.begin();

    // Compares all the entities for collisions, if a collision is found, then
    // it does a collision both ways
    while( entityIt != entities.end()-- ) {
        entityIt2 = ++entityIt;
        entityIt--;
        while( entityIt2 != entities.end() ) {
            if( (*entityIt)->checkCollision( *entityIt2 ) ) {

                (*entityIt)->doCollision( *entityIt2 );
                (*entityIt2)->doCollision( *entityIt );
            }
            ++entityIt2;
        }
        ++entityIt;
    }
}

/**
 * \todo map->draw: need to find viewport size instead of hardcoding.
 */
void Game::draw() {
    assert( map != NULL );

    for( entityIt = entities.begin(); entityIt != entities.end(); ++entityIt ) {
        map->addToRenderList( *(*entityIt) );
    }

    for ( effectIt = effects.begin(); effectIt != effects.end(); ++effectIt ) {
        map->addToRenderList( *(*effectIt) );
    }

    map->draw( 800, 600, Map::CAMERA0 | Map::CAMERA1 ); //hardcoded for now.
    map->clearRenderList();

    // Draw the HUD last
    Point pt1( 0.0f, 0.0f );
    player1->drawHUD( pt1 );
    Point pt2( (float)(( 800 / 2) + 5), 0.0f );
    player2->drawHUD( pt2 );
}

bool Game::control( float dt ) {
    assert( map != NULL );

    // Check the quit button and fullscreen/window toggle button
    if( g_key->isKeyDown(DIK_ESCAPE) ) {
        return true;
    }
    else if( g_key->isKeyDown(DIK_F1) ) {
        if( g_win->toggleWindowed() ) {
            TRACE( "Could not get Window." );
            assert( false );
            return true;
        }
    }

    // Check for the map loading keys
    bool isKeyDown = false;
    for( int i = 0; (i < 10) && !isKeyDown; i++ ) {
        isKeyDown = g_key->isKeyDown(DIK_1 + i);
        if( isKeyDown ) {
            if( i == 9 ) {
                keyPressed = 0;
            }
            else {
                keyPressed = i + 1;
            }
        }
    }
    // A number key was pressed and has been released
    // so start a new game using the map corresponsing to
    // the pressed key
    if( keyPressed != -1 && !isKeyDown ) {
        startNewGame( keyPressed );
        keyPressed = -1;
    }

    // Control all controllers
    for( ctrlIt = ctrls.begin(); ctrlIt != ctrls.end(); ++ctrlIt ) {
        (*ctrlIt)->control( dt );
    }    

    return false;
}

Map* Game::getMap() const {
    return map;
}

void Game::addEntity( Entity* e ) {
    entities.push_back(e);
}

void Game::addController( Controller* c ) {
    ctrls.push_back(c);
}

void Game::addLogic( GameLogic* g ) {
    logics.push_back(g);
}

void Game::addEffect( Effect* e ) {
    effects.push_back(e);
}
