/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#ifndef _GAME_H_
#define _GAME_H_

#include "KeyboardController.h"

class Entity;
class Controller;
class GameLogic;
class Map;
class Character;
class Effect;

/**
 * Contains the game objects, updates the logic, and draws
 * entities.
 */
class Game {

public:

    /**
     * Constructor
     */
    Game();

    /**
     * Destructor
     */
    ~Game();

    /**
     * Updates the game logic
     *
     * @param dt                the difference in time since last update
     * @returns                 true if it is time to exit the game
     */
    void update( float dt );

    /**
     * Draws all game objects
     */
    void draw();

    /**
     * Control all the game objects.
     *
     * @param                   the amount of time since the last frame
     * @returns                 true if it is time to exit the game
     */
    bool control( float dt );

    /**
     * Get the current map.
     *
     * @returns                 the current map
     */
    Map* getMap() const;

    /**
     * Add an entity to the game.  The entity will automatically be drawn
     * and updated every iteration of the game loop.
     *
     * @param e                 the entity to add
     */
    void addEntity( Entity* e );

    /**
     * Add a controller to the game.  Control will automatically be read
     * from the controller each iteration of the game loop.
     *
     * @param c                 the controller to add 
     */
    void addController( Controller* c );

    /**
     * Add a GameLogic to the game.  The object will be updated each iteration
     * of the game loop.
     *
     * @param g                 the logic to add
     */
    void addLogic( GameLogic* g );

    /**
     * Add an Effect to the game.  The object will be updated each iteration
     * of the game loop.
     *
     * @param e                 the Effect to add
     */
    void addEffect( Effect* e );

    /**
     * Starts a new game.  The parameter is the number of the map file in the
     * maps directory to use for the game (ex: 0 would load maps/map0.map).
     *
     * @param                   number of the map file
     * @returns                 true if the map load failed.
     */
    bool startNewGame( int mapNum );

    /**
     * Sets a flag so that the Map will restart the game completely anew at the
     * start of the next update call.
     */
    void resetMap();

private:

    /**
     * Deletes all entities, controllers, and the map so that another game
     * can start.
     */
    void clear();

private:

    /**
     * The entities currently in the game.
     */
    std::list<Entity *> entities;

    /**
     * The controllers currently used in the game.
     */
    std::list<Controller *> ctrls;

    /**
     * An iterator used to iterate through the entities.
     */
    std::list<Entity *>::iterator entityIt;

    /**
     * An iterator used to iterate through the controllers.
     */
    std::list<Controller *>::iterator ctrlIt;

    /**
     * List of GameLogics currently active in the game.
     */
    std::list<GameLogic *> logics;

    /**
     * An iterator used to iterate through the GameLogics.
     */
    std::list<GameLogic *>::iterator logicIt;

    /**
     * List of Effects currently active in the game.
     */
    std::list<Effect *> effects;

    /**
     * An iterator used to iterate through the Effects.
     */
    std::list<Effect *>::iterator effectIt;

    /**
     * The game map.
     */
    Map* map;

    /**
     * Control settings for player1
     */
    KeyboardSettings keys1;

    /**
     * Control settings for player2
     */
    KeyboardSettings keys2;

    /**
     * Which key is currently being pressed.  Used to start new games on key
     * release instead of on key down.
     */
    int keyPressed;

    /**
     * The first player's character
     */
    Character* player1;

    /**
     * The second player's character
     */
    Character* player2;

    /**
     * Number of the currently running map (as passed to startNewGame).
     */
    int currMap;

    /**
     * Flag set in response to the restartMap method.
     */
    bool resetMapFlag;

    /**
     * Number of times player1 has won
     */
    int numWins1;

    /**
     * Number of times player2 has won
     */
    int numWins2;

};

#endif
