/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#ifndef _GAMEDX_H_
#define _GAMEDX_H_

#include "IsoBomb.h"

class Window;
class GameDXListener;

/**
 * A container that holds and initializes all the DirectX objects
 */
class GameDX {

public:
    /**
     * Constructor
     *
     */
    GameDX();

    /**
     * Destructor
     *
     */
    ~GameDX();


public:
    /**
     * Sets up Direct3D
     *
     * @returns     the result
     */
    HRESULT setupD3D();

    /**
     * Sets up DirectInput
     *
     * @returns     the result
     */
    HRESULT setupDI();

    /**
     * Gets a handle to the Direct3D device
     *
     * @returns     a handle to the Direct3D device
     */    
    LPDIRECT3DDEVICE8 & getDeviceHandle();

    /**
     * Gets a reference to the Direct3D object
     *
     * @returns     a reference to the Direct3D object
     */
    LPDIRECT3D8 & getD3D();

    /**
     * Gets a reference to the client surface coordinates
     *
     * @returns     a reference to the client surface coordinates
     */
    D3DRECT & getClientSurfaceCoords();

    /**
     * Gets a reference to a DirectX Font object
     *
     * @returns     a reference to a DirectX Font object
     */
    LPD3DXFONT & getFont();

    /**
     * Gets a reference to the DirectInput object
     *
     * @returns     a reference to the DirectInput object
     */
    LPDIRECTINPUT8 & getDirectInput();

    /**
     * Releases all objects created for Direct3D
     *
     */
    void releaseD3D();

    /**
     * After setting up the D3D device, this resets the device, picking up the
     * windowed mode from the global window.  This works essentially the same
     * as setupD3D, except that it allows re-initialization after a setupD3D
     * call.
     *
     * D3DPOOL_MANAGED textures and objects will remain between resets, but
     * not between a release/setup call pair.
     */
    HRESULT resetD3D();

    /**
     * Releases all objects created for DirectInput
     *
     */
    void releaseDI();

    /**
     * Ensures that the listener given is registered.  If the given listener
     * has already registered, this method has no effect.
     */
    void registerListener( GameDXListener& listener );

    /**
     * Removes the given listener from the list of registered listeners, so
     * that it will no longer receive events.  If the given listener is not
     * in the list, this method will have no effect.
     */
    void removeListener( GameDXListener& listener );

private:
    /**
     * Sends lost event to all listeners.
     */
    void generateLostEvent();

    /**
     * Sends reset event to all listeners.
     */
    void generateResetEvent();

    /**
     * Initializes the d3dpp variable.  This method requires the m_pd3d object
     * to be properly set up.
     */
    HRESULT setupPresentParams( bool isWindowed );

private:

    /**
     * Direct3D display mode
     *
     */
    D3DDISPLAYMODE d3ddm;

    /**
     * Direct3D present parameters
     *
     */
    D3DPRESENT_PARAMETERS d3dpp;


    /**
     * Direct3D object
     *
     */
    LPDIRECT3D8 m_pd3d;

    /**
     * Direct3D device
     *
     */
    LPDIRECT3DDEVICE8 m_pd3dDevice;

    /**
     * Direct3D main surface
     *
     */
    LPDIRECT3DSURFACE8 m_pMainSurf;

    /**
     * Direct3D back surface
     *
     */
    LPDIRECT3DSURFACE8 m_pBackSurf;

    /**
     * Direct3D Font object
     *
     */
    LPD3DXFONT m_pd3dFont;

    /**
     * DirectInput object
     *
     */
    LPDIRECTINPUT8 m_pdi;

    typedef std::vector< GameDXListener* > LVector; //Listener Vector
    typedef LVector::iterator LVectorIter;          //Listener Vector iterator

    LVector listeners;

};

#endif
