/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#include "IsoBomb.h"
#include "HomingBomb.h"
#include "Map.h"
#include "Game.h"
#include "Tile.h"
#include "Explosion.h"
#include "SoundSystem.h"

HomingBomb::HomingBomb() : target(NULL), targetPos(-1.0, -1.0, 0.0, 0.0), hasPath(false) {
    sound = g_snd->startHomingBombSound();
}

HomingBomb::HomingBomb( const Rect& initPos, int initDir, int size, 
                       float delay, float initSpeed, const Character* c ) : 
    Bomb( initPos, initDir, (float)size, delay ), speed(initSpeed), 
    target(c), targetPos(-1.0, -1.0, 0.0, 0.0), hasPath(false), firstTime( true ) {

    anim.setOffsets( 20, 72 );
    feetAnim.setScaleFactor( 0.5f );
    feetAnim.setOffsets( 20, 40 );
    feetAnim.setAnimationFrames(g_am->getBombFeetAnimationFrames(DIRECTION_DOWNLEFT));
    feetAnim.setAnimRuntime( 1 / speed );

    sound = g_snd->startHomingBombSound();
}

HomingBomb::~HomingBomb() {
    assert( !g_snd->isTokenValid( sound ) );
}

bool HomingBomb::update( float dt ) {
	bool hasDied = Bomb::update( dt );

    // Update the height so that Bomb uses the right
    // height when create explosions.
    height = (float)g_game->getMap()->getRectHeight( pos );

    // Get initial path
    if( !hasPath ) {
        hasDied = !getNextDestination();
        hasPath = true;
    }

    if( !hasDied ) {
        move( dt );

        feetAnim.start();
        feetAnim.update( dt );

        if( hasReachedDestination() ) {
            hasDied = !getNextDestination();
        }
    }

    if( hasDied ) {
        blowUp();
        g_snd->stopHomingBombSound( sound );
    }

    return hasDied;	
}

void HomingBomb::draw( const Point& anchor ) {
    feetAnim.draw( anchor );
    Bomb::draw( anchor );
}

void HomingBomb::move( float dt ) {
    switch( direction ) {
    case DIRECTION_DOWNRIGHT:
        pos.x += (speed * dt);
        break;
    case DIRECTION_UPLEFT:
        pos.x -= (speed * dt);
        break;
    case DIRECTION_UPRIGHT:
        pos.y -= (speed * dt);
        break;
    case DIRECTION_DOWNLEFT:
        pos.y += (speed * dt);
        break;
    }

    keepInCenter();
    assert( !g_game->getMap()->isLocationOutside(pos) );
}

bool HomingBomb::hasReachedDestination() {
    Point center = pos.getCenter();
    switch( direction ) {
    case DIRECTION_DOWNRIGHT:
        return center.x > destPoint.x;
        break;
    case DIRECTION_UPLEFT:
        return center.x < destPoint.x;
        break;
    case DIRECTION_UPRIGHT:
        return center.y < destPoint.y;
        break;
    case DIRECTION_DOWNLEFT:
        return center.y > destPoint.y;
        break;
    }

    return false;
}

void HomingBomb::keepInCenter() {
    switch( direction ) {
    case DIRECTION_DOWNRIGHT:
    case DIRECTION_UPLEFT:
        pos.y = ((float)floor(pos.y) + 0.5f) - (pos.h / 2.0f);
        break;
    case DIRECTION_UPRIGHT:
    case DIRECTION_DOWNLEFT:
        pos.x = ((float)floor(pos.x) + 0.5f) - (pos.w / 2.0f);
        break;
    }
}

bool HomingBomb::getNextDestination() {
	bool destFound = true;

    // Test if the player's position has moved
    if( firstTime || g_game->getMap()->getTileAt(targetPos.getCenter()) !=
        g_game->getMap()->getTileAt(target->getPosition().getCenter()) ) {
        firstTime = false;

        targetPos = target->getPosition();
        if( finder.findpath(pos.getCenter(), targetPos.getCenter()) ) {
            if( !finder.hasMorePoints() ) {
                // The source and destination are the same, just blow up
                destFound = false;
            }
        }
        else {
            // No path found so just blow up
            destFound = false;
        }		
	}
	else {
		if( !finder.hasMorePoints() ) {
			destFound = false;
		}
	}

	if( destFound ) {
		destPoint = finder.getNextPoint();
   
		Point center = pos.getCenter();
		int posX = (int)center.x;
		int posY = (int)center.y;
		int destX = (int)destPoint.x;
		int destY = (int)destPoint.y;

		// Figure what direction to move
		if( posX < destX ) {
			anim.setAnimationFrames(g_am->getBombAnimationFrames(DIRECTION_DOWNRIGHT));
			feetAnim.setAnimationFrames(g_am->getBombFeetAnimationFrames(DIRECTION_DOWNRIGHT));
			direction = DIRECTION_DOWNRIGHT;
		}
		else if( posX > destX ) {
			anim.setAnimationFrames(g_am->getBombAnimationFrames(DIRECTION_UPLEFT));
			feetAnim.setAnimationFrames(g_am->getBombFeetAnimationFrames(DIRECTION_UPLEFT));
			direction = DIRECTION_UPLEFT;
		}
		else if( posY < destY ) {
			anim.setAnimationFrames(g_am->getBombAnimationFrames(DIRECTION_DOWNLEFT));
			feetAnim.setAnimationFrames(g_am->getBombFeetAnimationFrames(DIRECTION_DOWNLEFT));
			direction = DIRECTION_DOWNLEFT;
		}
		else if( posY > destY ) {
			anim.setAnimationFrames(g_am->getBombAnimationFrames(DIRECTION_UPRIGHT));
			feetAnim.setAnimationFrames(g_am->getBombFeetAnimationFrames(DIRECTION_UPRIGHT));
			direction = DIRECTION_UPRIGHT;
		}
	}

	return destFound;
}
