#include "IsoBomb.h"

/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#include "Game.h"
#include "AnimationManager.h"
#include "TextureManager.h"
#include "IsoException.h"
#include "SoundSystem.h"
#include "ConfigFile.h"
#include "Powerup.h"
#include "EndGame.h"
#include "Particle.h"
#include "Virus.h"

GameDX* g_dx = NULL;
Keyboard* g_key = NULL;
Window* g_win = NULL;
Game* g_game = NULL;
AnimationManager* g_am = NULL;
TextureManager* g_tm = NULL;
SoundSystem* g_snd = NULL;
ConfigFile* g_cfg = NULL;
TextureManager* g_hudTm = NULL;

int g_winSizeX = NULL;
int g_winSizeY = NULL;
static bool s_isPaused = false;

void initializeGame( HINSTANCE hInstance, int nShowCmd );
int messageLoop();

int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
                      LPSTR lpCmdLine, int nShowCmd) {

    int ret = 0;

    try {
        initializeGame( hInstance, nShowCmd );
        ret = messageLoop();

    } catch ( const IsoException& e ) {
        e.showMessageBox();
    }

    Particle::releaseTextures();
    delete g_game;
    delete g_key;
    delete g_am;
    delete g_tm;
    delete g_hudTm;
    delete g_snd;
    delete g_dx;
    delete g_win;
    delete g_cfg;

    return ret;
}

void initializeGame( HINSTANCE hInstance, int nShowCmd ) {
    HRESULT r;

    // Initialise the random number generators
    srand ( time(NULL) );
    randomize(); //init JRND

    // Setup configuration file
    g_cfg = new ConfigFile( "isobomb.cfg" );
    Powerup::initVars( *g_cfg );
    EndGame::initTimeToReset( *g_cfg );
    Character::initVars( *g_cfg );
    Virus::initVars( *g_cfg );

    // Setup Window
    g_win = new Window;
    g_winSizeX = DEFAULT_WINDOW_WIDTH;
    g_winSizeY = DEFAULT_WINDOW_HEIGHT;
    int fullscreen = 0;
    g_cfg->getVar( "", "fullscreen", fullscreen, 0 );
    g_win->setWindowed( fullscreen == 0 );
    g_win->registerClass( hInstance, WINDOW_TITLE );
    if( !g_win->initInstance( hInstance, nShowCmd, WINDOW_TITLE, WINDOW_TITLE) ) {
        //I do the delete here, because IsoException uses g_win to generate the
        //MessageBox, so g_win has to have a valid HWND.
        delete g_win;
        g_win = NULL;
        throw IsoException( "Error Creating Window" );
    }

    // Setup DirectX and tie it to our window
    g_dx = new GameDX;
    r = g_dx->setupD3D();
    if( FAILED(r) ) {
        throw IsoException("Error Setting Up Direct3D");
    }
    // Setup DirectInput
    r = g_dx->setupDI();
    if( FAILED(r) ) {
        throw IsoException("Error Setting Up DirectInput");
    }

    // Initialize Sound
    g_snd = new SoundSystem();

    // Initialize keyboard
    g_key = new Keyboard;
    g_key->initialize( g_win->getHwnd() );

    // Initialize graphics
    g_am = new AnimationManager;
    g_tm = new TextureManager( "tiles", "bmp", g_dx->getDeviceHandle() );
    g_hudTm = new TextureManager( "grfx/HUD", "png", g_dx->getDeviceHandle() );
    Particle::initTextures( g_dx->getDeviceHandle() );

    // Create the Game object
    g_game = new Game;

    // Start the game
    int startingMap;
    g_cfg->getVar( "Game", "startingMap", startingMap, 0 );
    if ( g_game->startNewGame( startingMap ) )
        throw IsoException( "Error loading the starting map." );
}

int messageLoop() {
    // Intialize timer
    Timer timer;
    timer.setStartTime();

    float dt;

    MSG msg;
    while(TRUE) {
        if ( !s_isPaused ) {
            //check and see if there is a message
            if (PeekMessage( &msg, NULL, 0, 0, PM_REMOVE)) {
                //check and see if this is the quit message
                if (msg.message == WM_QUIT)
                    //exit the message loop
                    break;
                //else do what we would have before, namely translate and dispatch
                TranslateMessage( &msg );
                DispatchMessage( &msg );
            }
            else {
                HRESULT r = 0;
                
                dt = timer.getFrameDt();

                // Control the game logic returns true when it's time to exit
                if( g_game->control(dt) ) {
                    break;
                }

                g_game->update( dt );

                r = g_dx->getDeviceHandle()->BeginScene();
                if ( SUCCEEDED( r ) ) {
                    // Draw all entities
                    g_game->draw();

                    // Display frame rate
                    if( DISPLAY_FPS ) {
                        RECT rect =  {20, 20, 200, 200};
                        //char frateint[32];
                        char framerate[32];
                        //ZeroMemory( &frateint, sizeof(frateint) );
                        //strcpy( framerate, "Frame Rate: ");
                        itoa( timer.getFrameRate(), framerate, 10);
                        //strcat( framerate, frateint );
                        g_dx->getFont()->DrawText( framerate, -1, &rect, DT_NOCLIP, D3DCOLOR_XRGB(255,255,255) );
                    }

                    g_dx->getDeviceHandle()->EndScene();
                    g_dx->getDeviceHandle()->Present( NULL, NULL, NULL, NULL );

                    timer.calcFrameCount();
                } else {
                    TRACE( "BeginScene call failed!" );
                }
            }

        } else {
            //We want to wait for messages when paused.
            GetMessage( &msg, NULL, 0, 0 );
            //check and see if this is the quit message
            if (msg.message == WM_QUIT) {
                //exit the message loop
                break;
            }
            //else do what we would have before, namely translate and dispatch
            TranslateMessage( &msg );
            DispatchMessage( &msg );
        }
    }

    return (int)msg.wParam;
}

//Global WndProc (Can't be placed inside class :<  )
LRESULT CALLBACK GlobalWndProc( HWND hWnd, UINT message, 
							    WPARAM wParam, LPARAM lParam ) {

    HDC hdc;
    PAINTSTRUCT ps; //structure used by Windows to paint things
    switch(message) {

        case WM_CREATE:
            return DefWindowProc( hWnd, message, wParam, lParam );
            break;
        case WM_PAINT:
            hdc = BeginPaint(hWnd, &ps);
            //do any windows drawing here
            EndPaint(hWnd, &ps);
            break;
        case WM_CLOSE:
            TRACE("Posting Destroy");
            PostQuitMessage(0);
            break;
        case WM_SIZE:
            if ( wParam == SIZE_RESTORED || wParam == SIZE_MAXIMIZED ) {
                // Reset window size when size is changed
                g_winSizeX = LOWORD(lParam);
                g_winSizeY = HIWORD(lParam);
                s_isPaused = false;
            } else if ( wParam == SIZE_MINIMIZED ) {
                s_isPaused = true;
            }
            break;
        case WM_ACTIVATE:
            if( (LOWORD(wParam) != WA_INACTIVE) && (HIWORD(wParam) == 0) ) {
                // If window loses focus, need to reinitialize keyboard
                if( g_key != NULL ) {
                    g_key->initialize( hWnd );
                }
            }
            break;
        case WM_SETCURSOR:	
            if( !g_win->getWindowed() ) {
                SetCursor( NULL );
                if( g_dx->getDeviceHandle() != NULL ) {
                    g_dx->getDeviceHandle()->ShowCursor( FALSE );
                }
            }
            else {
                return DefWindowProc( hWnd, message, wParam, lParam );
            }
            break;
        default:
            return DefWindowProc(hWnd, message,wParam,lParam);
    }

    return 0;
}

void doTrace( const char* fn, int lineno, const char* msg, ... ) {
    va_list arg;
    va_start( arg, msg );
    char buf[ 1024 ];
    vsprintf( buf, msg, arg );
    va_end( arg );

    char buf2[ 1024 ];
    sprintf( buf2, "File: %s, Line: %i: %s\n", fn, lineno, buf );

    OutputDebugString( buf2 );
}
