/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#include "IsoBomb.h"
#include "Map.h"
#include "TileLocation.h"
#include "TextureManager.h"
#include "Tile.h"
#include "Point.h"
#include "Point3D.h"
#include "Entity.h"
#include "Powerup.h"
#include "Game.h"

//Used for temporary in-row ordering.
struct TileDrawInfo {
    int tx; //tile coordinates
    int ty; 
    int sx; //Screen coordinates
    int sy;
    int h;  //Tile height

    /**
     * Operator that sorts by tile height.
     */
    bool operator >= ( const TileDrawInfo& rhs ) const {
        return ( h >= rhs.h );
    }
};

typedef std::list< TileDrawInfo > TDList;
typedef TDList::iterator TDListIter;

//How tall and wide tiles are when drawn on the screen.
const int TILE_SCREEN_W = 128;
const int TILE_SCREEN_H = 64;
//The height on the screen of one logical height unit.
const int HEIGHT_SCREEN_H = 16;
//What % of the texture takes up each height unit?
const float TEX_PER_HEIGHT = 0.25f;

//The expected max tile height in pixels, so we know how far below the screen to draw.
const int MAX_TILE_HEIGHT = 8 * HEIGHT_SCREEN_H;

//size of the divider in split screen
const int DIVIDER_SIZE = 4;

#define MAP_MAX(x, y) ((x > y) ? (x) : (y))

//Flexible vertex format for the TileVertex struct.
//If you move this line, edit Map.h's comment for TileVertex saying the FVF
//flags are defined in Map.cpp and not in the header.
const DWORD TileVertexFVF = ( D3DFVF_XYZRHW | D3DFVF_TEX1 );

const int Map::CAMERA0 = 1;
const int Map::CAMERA1 = 2;

Map::Map() : mapLength(0), mapWidth(0), tiles(NULL) {
    cameraTarget[0] = cameraTarget[1] = NULL;

    assert( g_dx != NULL );
    //Textures always remain the same.
    vertex[0].u = 0.0f;
    vertex[0].v = 1.0f;
    vertex[1].u = 0.0f;
    vertex[1].v = 0.0f;
    vertex[2].u = 1.0f;
    vertex[2].v = 1.0f;
    vertex[3].u = 1.0f;
    vertex[3].v = 0.0f;

    //We aren't using the Z axis.
    vertex[0].z = 0.0f;
    vertex[0].rhw = 1.0f;
    vertex[1].z = 0.0f;
    vertex[1].rhw = 1.0f;
    vertex[2].z = 0.0f;
    vertex[2].rhw = 1.0f;
    vertex[3].z = 0.0f;
    vertex[3].rhw = 1.0f;

    //Now initialize the side vextex drawing, points in the strip as below:
    //1   5
    //** **
    //**3**
    //*****
    //0***4
    // ***
    //  2
    sideVertex[0].u = 0.0f;
    sideVertex[0].v = 1.0f;
    sideVertex[0].z = 0.0f;
    sideVertex[0].rhw = 1.0f;

    sideVertex[1].u = 0.0f;
    sideVertex[1].v = 0.0f;
    sideVertex[1].z = 0.0f;
    sideVertex[1].rhw = 1.0f;

    sideVertex[2].u = 1.0f;
    sideVertex[2].v = 1.0f;
    sideVertex[2].z = 0.0f;
    sideVertex[2].rhw = 1.0f;

    sideVertex[3].u = 1.0f;
    sideVertex[3].v = 0.0f;
    sideVertex[3].z = 0.0f;
    sideVertex[3].rhw = 1.0f;

    sideVertex[4].u = 2.0f;
    sideVertex[4].v = 1.0f;
    sideVertex[4].z = 0.0f;
    sideVertex[4].rhw = 1.0f;

    sideVertex[5].u = 2.0f;
    sideVertex[5].v = 0.0f;
    sideVertex[5].z = 0.0f;
    sideVertex[5].rhw = 1.0f;
}

Map::~Map() {
    if( tiles != NULL ) {
        for( int i = 0; i < mapLength; i++ ) {
            delete[] tiles[i];
        }
        delete[] tiles;
    }
}

void Map::render( const Point &pt, int scrWidth, int scrHeight, int xShift ) {
    assert( scrWidth > 0 && scrHeight > 0 );
    g_dx->getDeviceHandle()->Clear( 0, 0, D3DCLEAR_TARGET, D3DCOLOR_XRGB(64, 0, 0), 0.0f, 0 );

    //Get this point's offset, used to adjust screenX and screenY later for a
    //smooth scrolling effect.
    Point offset = getScreenTileOffset( pt, floorf(pt.x), floorf(pt.y) );

    //Find the tile and screen location in the center.
    TileLocation walker( (int)pt.x, (int)pt.y );
    int screenX( scrWidth / 2 - (int)offset.x );
    int screenY( scrHeight / 2 - (int)offset.y );

    //Search for the upper left tile.  We could probably optimize this later
    //by doing some integer division and modulus math.
    while ( screenY + TILE_SCREEN_H/2 > 0 ) {
        walker.north();
        screenY -= TILE_SCREEN_H;
    }
    while ( screenX + TILE_SCREEN_W/2 > 0 ) {
        walker.west();
        screenX -= TILE_SCREEN_W;
    }

    //Now draw out a rectangular patch region of map
    TileLocation rowStart( walker );
    int screenStartX = screenX;

    int rowsDrawn = 0;
    while ( screenY - TILE_SCREEN_H/2 < scrHeight + MAX_TILE_HEIGHT ) {
        ++rowsDrawn;

        //Order the tiles on this row for drawing
        TDList drawList;
        drawList.clear();
        while ( screenX - TILE_SCREEN_W/2 < scrWidth ) {
            if ( !isLocationOutside( walker ) ) {
                TileDrawInfo temp;
                temp.tx = walker.getX();
                temp.ty = walker.getY();
                temp.sx = screenX;
                temp.sy = screenY;
                temp.h = tiles[temp.ty][temp.tx].height;

                //Sorted insert
                drawList.insert(
                    find_if( drawList.begin(), drawList.end(),
                        bind2nd( greater_equal<TileDrawInfo>(), temp ) ), temp );
            }

            //Continue rendering the row horizontally.
            walker.east();
            screenX += TILE_SCREEN_W;
        }

        //Now do the actual drawing
        TDListIter iter;
        int lastH = -1;
        for ( iter = drawList.begin(); iter != drawList.end(); ++iter ) {
            TileDrawInfo& tdi = (*iter);
            assert( tdi.h >= lastH );
            lastH = tdi.h;

            drawTile( tdi.tx, tdi.ty, tdi.sx, tdi.sy, xShift );
            drawEntitiesOnTile( tdi.tx, tdi.ty, tdi.sx, tdi.sy, xShift );
            //g_dx->getDeviceHandle()->EndScene();
            //g_dx->getDeviceHandle()->Present( NULL, NULL, NULL, NULL );
            //g_dx->getDeviceHandle()->BeginScene();
        }

        //This moves us down "half" a row.  We have to alternate our SE/SW
        //movements to make sure we fill the entire screen with tiles.
        if ( screenStartX < -TILE_SCREEN_W/2 ) {
            rowStart.southeast();
            screenStartX += TILE_SCREEN_W/2;
        } else {
            rowStart.southwest();
            screenStartX -= TILE_SCREEN_W/2;
        }
        screenY += TILE_SCREEN_H/2;
        walker = rowStart;
        screenX = screenStartX;
    }
}

void Map::draw( const Point& pt, int scrWidth, int scrHeight ) {
    D3DVIEWPORT8 vp;
    vp.X = 0;
    vp.Y = 0;
    vp.Height = scrHeight;
    vp.Width = scrWidth;
    vp.MinZ = 0.0f;
    vp.MaxZ = 1.0f;

    int r = g_dx->getDeviceHandle()->SetViewport( &vp );
    assert( SUCCEEDED( r ) );
    render( pt, vp.Width, vp.Height, vp.X );
}

void Map::draw( int scrWidth, int scrHeight, int cameraFlags ) {
    assert( ( cameraFlags & ~( CAMERA0 | CAMERA1 ) ) == 0 );

    D3DVIEWPORT8 oldViewport;
    int hr = 0;
    hr = g_dx->getDeviceHandle()->GetViewport( &oldViewport );
    assert( SUCCEEDED( hr ) );

    //set up a default full-screen viewport.
    D3DVIEWPORT8 vp;
    vp.X = 0;
    vp.Y = 0;
    vp.Height = scrHeight;
    vp.Width = scrWidth;
    vp.MinZ = 0.0f;
    vp.MaxZ = 1.0f;

    if ( cameraFlags == 3 ) {
        assert( cameraTarget[0] != NULL && cameraTarget[1] != NULL );

        //Both are set, so we need to do split screen
        //render left half
        vp.Width = scrWidth / 2 - DIVIDER_SIZE / 2;
        hr = g_dx->getDeviceHandle()->SetViewport( &vp );
        assert( SUCCEEDED( hr ) );
        render( cameraTarget[0]->getPosition().getCenter(), vp.Width, vp.Height, vp.X );

        //render right half
        vp.X = scrWidth / 2 + DIVIDER_SIZE / 2;
        hr = g_dx->getDeviceHandle()->SetViewport( &vp );
        assert( SUCCEEDED( hr ) );
        render( cameraTarget[1]->getPosition().getCenter(), vp.Width, vp.Height, vp.X );

        //clear the center
        vp.X = vp.Width;
        vp.Width = DIVIDER_SIZE;
        hr = g_dx->getDeviceHandle()->SetViewport( &vp );
        assert( SUCCEEDED( hr ) );
        hr = g_dx->getDeviceHandle()->Clear( 0, 0, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 0.0f, 0 );
        assert( SUCCEEDED( hr ) );

    } else {
        int renderCam = 0;
        if ( cameraFlags & CAMERA0 )
            renderCam = 0;
        else if ( cameraFlags & CAMERA1 )
            renderCam = 1;

        assert( cameraTarget[ renderCam ] != NULL );

        //We can use the default full-screen viewport
        hr = g_dx->getDeviceHandle()->SetViewport( &vp );
        assert( SUCCEEDED( hr ) );
        render( cameraTarget[ renderCam ]->getPosition().getCenter(), vp.Width, vp.Height, vp.X );
    }

    //restore old viewport
    hr = g_dx->getDeviceHandle()->SetViewport( &oldViewport );
    assert( SUCCEEDED( hr ) );
}

void Map::setCameraTarget( const Entity* target, int camera ) {
    assert( camera == 0 || camera == 1 );

    cameraTarget[camera] = target;
}

const Tile* Map::getTileAt( const Point &pt ) const {
    assert( !isLocationOutside( pt ) );
    //if ( !isLocationOutside( pt ) )
        return &( tiles[ (int)pt.y ][ (int)pt.x ] );
    //else
    //    return NULL;
}

bool Map::isLocationOutside( const TileLocation& loc ) const {
    return ( loc.getX() < 0 || loc.getY() < 0 ||
             loc.getX() >= mapWidth || loc.getY() >= mapLength );
}

bool Map::isLocationOutside( const Point& loc ) const {
    return ( loc.x < 0 || loc.y < 0 || loc.x >= mapWidth || loc.y >= mapLength );
}

bool Map::isLocationOutside( const Point3D& loc ) const {
    return ( loc.x < 0 || loc.y < 0 || loc.x >= mapWidth || loc.y >= mapLength );
}

bool Map::isLocationOutside( const Rect& r ) const {
    return isLocationOutside( r.getUpperLeft() ) || isLocationOutside( r.getLowerRight() );
}

int Map::getRectHeight( const Rect& r ) const {
    return MAP_MAX( getTileAt( r.getUpperLeft() )->height, 
           MAP_MAX( getTileAt( r.getUpperRight() )->height,
           MAP_MAX( getTileAt( r.getLowerRight() )->height,
                    getTileAt( r.getLowerLeft() )->height ) ) );
}

int Map::getPointHeight( const Point& p ) const {
    assert( !isLocationOutside( p ) );
    return getTileAt( p )->height;
}

int Map::getPointHeight( const Point3D& p ) const {
    assert( !isLocationOutside( p ) );
    return ( tiles[ (int)p.y ][ (int)p.x ].height );
}

void Map::addToRenderList( Drawable& obj ) {
    using namespace std;

    //Get objects's anchor
    Rect r = obj.getPosition();
    Point p = r.getLowerRight();
    assert( !isLocationOutside( p ) );

    //Get tile coordinates
    int y = (int)p.y;
    int x = (int)p.x;

    ObjListNode newNode;
    newNode.obj = &obj;
    newNode.scrOffset = getScreenTileOffset( r.getCenter(), (float)x, (float)y );

    Tile::DrawListIter iter =
        find_if( tiles[y][x].objs.begin(), tiles[y][x].objs.end(),
                 bind2nd( greater_equal<ObjListNode>(), newNode ) );

    tiles[y][x].objs.insert( iter, newNode );
}

void Map::clearRenderList() {
    for ( int y = 0; y < mapLength; ++y )
        for ( int x = 0; x < mapWidth; ++x )
            tiles[y][x].objs.clear();
}

Point Map::getScreenTileOffset( const Point& pt, float startX, float startY ) {
    //The ratio is 2x:1y for our isometric projection, this is already reflected
    //in the tile screen constants.

    float ptx = pt.x - startX;
    float pty = pt.y - startY;

    //"Follow" the x axis:
    float x = TILE_SCREEN_W/2 * ptx - 0;
    float y = TILE_SCREEN_H/2 * ptx - TILE_SCREEN_H/2;
    //we added 0 and TILE_SCREEN_H/2 since we are finding from center of the tile.

    //"Follow" the y axis:
    x -= TILE_SCREEN_W/2 * pty;
    y += TILE_SCREEN_H/2 * pty;

    return Point( x, y );
}

bool Map::load( const char *fn ) {
    //We assert this since we don't properly clean up on multiple calls on
    //load.
    assert( tiles == NULL );

    std::fstream filestr;

    filestr.open( fn, std::fstream::in );

    if( !filestr.is_open() ) {
        TRACE1( "File (%s) cannot be opened", fn );
        return false;
    }
    else {
        TRACE1( "Loading map file (%s)", fn );

        char line[256];
        int lineNum = 0;
        int row = 0;
        int col = 0;

        while( filestr.good() ) {
            // Stop reading if we got enough tiles
            if( (row >= mapLength) && (lineNum != 0) ) {
                break;
            }

            filestr.getline( line, sizeof(line) );
            
            // Use stringstream to tokenize lines
            std::stringstream stringstr( line, std::stringstream::in );

            while( stringstr.good() ) {
                lineNum++;

                // First line contains map length and width
                if( lineNum == 1 ) {
                    stringstr >> mapLength;
                    stringstr >> mapWidth;

                    // Error if values for length and width or wrong or
                    // there is more input
                    if( (mapLength <= 0) || (mapWidth <= 0) || !stringstr.eof() ) {
                        TRACE1( "Map file format error (line %d): invalid or too many values", lineNum );
                        filestr.close();
                        return false;                        
                    }

                    TRACE2( "length=%d,width=%d", mapLength, mapWidth );

                    // Create tiles array
                    tiles = new Tile*[mapLength];
                    for( int i = 0; i < mapLength; i++ ) {
                        tiles[i] = new Tile[mapWidth];
                    }
                }
                else {
                    char word[16];
                    stringstr >> word;

                    // If first word is a #, it is a comment line
                    if( strcmp(word, "#") == 0 ) {
                        break;
                    }
                    else {
                        int height, topTex, sideTex;
                        char item;
                        height = atoi( word );
                        stringstr >> topTex;
                        stringstr >> sideTex;
                        
                        // Get the item if there is one specified
                        if( !stringstr.eof() ) {
                            stringstr>> item;
                            // Make sure to get all data on this line.
                            // Without this call, the next line will be skipped.
                            int tmp;
                            stringstr >> tmp;
                        }
                        else {
                            // Hyphen means no item specified
                            item = '-';
                        }

                        // Error if any of the values are not within the valid range
                        if( (height < 0) || (topTex < 0) || (sideTex < 0) ) {
                            TRACE1( "Map file format error (line %d): invalid value", lineNum );
                            filestr.close();
                            return false;                  
                        }

                        // Setup the tile
                        setTileAt( row, col, height, topTex, sideTex );

                        // Process the item (if there is one)
                        if( item != '-' ) {
                            Rect itemPos( (float)col, (float)row, 0.99f, 0.99f );
                            Powerup* pow;

                            switch( item ) {
                            // Starting point
                            case 'S':
                                // Place starting point in the center of the tile.
                                startPoints.push_back( itemPos.getCenter() );
                                break;
                            // Powerup types
                            case HP_UP:
                            case SPEED_UP:
                            case BOMBS_UP:
                            case BLAST_UP:
                            case THROW_UP:
                            case HOMING_UP:
                                pow = new Powerup( itemPos, (float)height, item );
                                powerups.push_back( pow );
                                break;
                            default:
                                TRACE1( "Map file format error (line %d): invalid item", lineNum );
                            }
                        }
                        
                        // Update map position
                        col++;
                        if( col == mapWidth ) {
                            row++;
                            col = 0;
                        }
                    }
                }
            }
        }

        // Error if there weren't enough rows specified in the file
        if( row != mapLength ) {
            TRACE( "Map file format error: insufficient tiles" );
            filestr.close();
            return false;
        }
    }

    // If it made it this far, it's all good
    filestr.close();
    return true;
}

void Map::setTileAt( int row, int col, int height, int topTex, int sideTex ) {
    assert( row >= 0 && row < mapLength );
    assert( col >= 0 && col < mapWidth );
    assert( tiles != NULL );

    tiles[row][col].height = height;
    tiles[row][col].topTex = g_tm->getTexture( topTex );
    tiles[row][col].sideTex = g_tm->getTexture( sideTex );
}

void Map::drawTile( int mapX, int mapY, int screenX, int screenY, int xShift ) {
    //Our vertices are unprojected, so we have to compensate for the viewport.
    screenX += xShift;

    //Discover the height for the tile
    int height = tiles[ mapY ][ mapX ].height * HEIGHT_SCREEN_H;

    //Vertex 0 is the "West" corner, 1 North, 2 South, 3, East.
    vertex[0].x = (float)(screenX - TILE_SCREEN_W/2);
    vertex[0].y = (float)(screenY - height);
    vertex[1].x = (float)screenX;
    vertex[1].y = (float)(screenY - TILE_SCREEN_H/2 - height);
    vertex[2].x = (float)screenX;
    vertex[2].y = (float)(screenY + TILE_SCREEN_H/2 - height);
    vertex[3].x = (float)(screenX + TILE_SCREEN_W/2);
    vertex[3].y = (float)(screenY - height);

    LPDIRECT3DDEVICE8& d = g_dx->getDeviceHandle();
    HRESULT r = 0;
    //DWORD oldShader;

    //r = d->GetVertexShader( &oldShader );
    //assert( SUCCEEDED( r ) );

    r = d->SetVertexShader( TileVertexFVF );
    assert( SUCCEEDED( r ) );

    r = d->SetTexture( 0, tiles[ mapY ][ mapX ].topTex );
    assert( SUCCEEDED( r ) );

    r = d->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, vertex, sizeof( vertex[0] ) );
    assert( SUCCEEDED( r ) );

    //Now draw the "side" of the tile
    if ( height > 0 ) {
        float vAdjust = 1.0f - (float)tiles[ mapY ][ mapX ].height * TEX_PER_HEIGHT;

        sideVertex[0].x = vertex[0].x;
        sideVertex[0].y = (float)screenY;

        sideVertex[1].v = vAdjust;
        sideVertex[1].x = vertex[0].x;
        sideVertex[1].y = vertex[0].y;

        sideVertex[2].x = vertex[2].x;
        sideVertex[2].y = (float)(screenY + TILE_SCREEN_H/2);

        sideVertex[3].v = vAdjust;
        sideVertex[3].x = vertex[2].x;
        sideVertex[3].y = vertex[2].y;

        sideVertex[4].x = vertex[3].x;
        sideVertex[4].y = (float)screenY;

        sideVertex[5].v = vAdjust;
        sideVertex[5].x = vertex[3].x;
        sideVertex[5].y = vertex[3].y;

        r = d->SetTexture( 0, tiles[ mapY ][ mapX ].sideTex );
        assert( SUCCEEDED( r ) );

        r = d->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 4, sideVertex, sizeof( sideVertex[0] ) );
        assert( SUCCEEDED( r ) );
    }

    //r = d->SetVertexShader( oldShader );
    //assert( SUCCEEDED( r ) );
}

void Map::drawBoundingBox( const Rect& r, int mapX, int mapY, int screenX, int screenY ) {
    Point currPoint = getScreenTileOffset( r.getLowerLeft(), (float)mapX, (float)mapY );
    vertex[0].x = currPoint.x + screenX;
    vertex[0].y = currPoint.y + screenY;
    vertex[4] = vertex[0];

    currPoint = getScreenTileOffset( r.getUpperLeft(), (float)mapX, (float)mapY );
    vertex[1].x = currPoint.x + screenX;
    vertex[1].y = currPoint.y + screenY;

    currPoint = getScreenTileOffset( r.getUpperRight(), (float)mapX, (float)mapY );
    vertex[2].x = currPoint.x + screenX;
    vertex[2].y = currPoint.y + screenY;

    currPoint = getScreenTileOffset( r.getLowerRight(), (float)mapX, (float)mapY );
    vertex[3].x = currPoint.x + screenX;
    vertex[3].y = currPoint.y + screenY;

    HRESULT result = 0;
    LPDIRECT3DDEVICE8& d = g_dx->getDeviceHandle();
    DWORD oldShader;

    result = d->GetVertexShader( &oldShader );
    assert( SUCCEEDED( result ) );

    result = d->SetVertexShader( TileVertexFVF );
    assert( SUCCEEDED( result ) );

    result = d->SetTexture( 0, NULL );
    assert( SUCCEEDED( result ) );

    result = d->DrawPrimitiveUP( D3DPT_LINESTRIP, 4, vertex, sizeof( vertex[0] ) );
    assert( SUCCEEDED( result ) );

    result = d->SetVertexShader( oldShader );
    //assert( SUCCEEDED( result ) );
}

void Map::drawEntitiesOnTile( int mapX, int mapY, int screenX, int screenY, int xShift ) {
    Tile::DrawList& objs = tiles[mapY][mapX].objs;

    if ( objs.size() > 0 ) {
        Tile::DrawListIter iter = objs.begin();
        while ( iter != objs.end() ) {
            const Rect& objPos = (*iter).obj->getPosition();
            float objHeight = (*iter).obj->getHeight();
            if ( objHeight < 0.0f )
                objHeight = (float)(getRectHeight( objPos ) * HEIGHT_SCREEN_H);
            else
                objHeight *= (float)HEIGHT_SCREEN_H;

            Point anchor = (*iter).scrOffset;
            anchor.x += screenX;
            if ( (*iter).obj->needsTransform() )
                anchor.x += (float)xShift;
            anchor.y += (float)screenY - objHeight;

            //Draw bounding box for debugging purposes.
            if( DRAW_BOUNDING_BOXES ) {
                drawBoundingBox( objPos, mapX, mapY, screenX + xShift, screenY );
                if ( objHeight > 0 )
                    drawBoundingBox( objPos, mapX, mapY, screenX + xShift, (int)(screenY - objHeight) );
            }


            (*iter).obj->draw( anchor );
            ++iter;
        }
    }
}

int Map::getMapLength() const {
    return mapLength;
}

int Map::getMapWidth() const {
    return mapWidth;
}

Point Map::getStartingPoint() {
    assert( !startPoints.empty() );

    int ind = rand() % startPoints.size();
    Point pt = startPoints[ind];
    startPoints.erase( startPoints.begin() + ind );
    return pt;
}

void Map::addPowerups() {
    std::list<Powerup*>::iterator powIt;
    for( powIt = powerups.begin(); powIt != powerups.end(); ++powIt ) {
        g_game->addEntity( (*powIt) );
    }
}