/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#include "IsoBomb.h"
#include "IsoException.h"
#include "Particle.h"
#include "Rect.h"
#include "Point3D.h"
#include "Vector3D.h"
#include "Map.h"
#include "Game.h"

const DWORD ParticleVertexFVF = ( D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1 );

//Gravity constant in vertical units per second per second.
const float GRAVITY = 15.0f;
const float DEF_BOUNCE_FACTOR = -0.4f;
const float DEF_SIZE = 3.0f;

Particle::ParticleVertex Particle::v[4];

Texture Particle::tex = Texture( 0 );

Particle::Particle( const Point3D& initPos, const Vector3D& initVect, float life )
: life( life ), pos( initPos ), vect( initVect ),
cached( initPos.x, initPos.y, 0.0f, 0.0f ), color( D3DCOLOR_XRGB( 255, 255, 255 ) ),
size( DEF_SIZE ), bounceF( DEF_BOUNCE_FACTOR ) {
}

Particle::~Particle() {
}

void Particle::initTextures( GraphicsDevice dev ) {
    /*
    if ( FAILED( D3DXCreateTextureFromFile( dev, "grfx/particles/0.png", &(tex) ) ) )
        throw IsoException( "Particle: Failed to load grfx/particles/0.png" );
        */

    v[0].z = 0.0f;
    v[0].rhw = 1.0f;
    v[0].u = 0.0f;
    v[0].v = 0.0f;

    v[1].z = 0.0f;
    v[1].rhw = 1.0f;
    v[1].u = 1.0f;
    v[1].v = 0.0f;

    v[2].z = 0.0f;
    v[2].rhw = 1.0f;
    v[2].u = 0.0f;
    v[2].v = 1.0f;

    v[3].z = 0.0f;
    v[3].rhw = 1.0f;
    v[3].u = 1.0f;
    v[3].v = 1.0f;
}

void Particle::releaseTextures() {
    if ( tex != 0 ) {
        tex->Release();
        tex = 0;
    }
}

void Particle::setColor( DWORD color ) {
    this->color = color;
}

void Particle::setSize( float size ) {
    this->size = size;
}

void Particle::setBounceFactor( float fact ) {
    bounceF = -fact;
}

bool Particle::update( float dt ) {
    //Add the effects of gravity to our vector
    Vector3D gravVect( 0.0f, 0.0f, -GRAVITY * dt );

    vect += gravVect;

    //Project our point along our movement vector for this frame.
    Vector3D fracVect( vect );
    fracVect.mult( dt );

    Point3D oldPos( pos );
    fracVect.addTo( pos );

    if ( !g_game->getMap()->isLocationOutside( pos ) ) {
        if ( (float)g_game->getMap()->getPointHeight( pos ) >= pos.z ) {
            //Did we bounce off the side of a wall or the floor?
            if ( (int)oldPos.x != (int)pos.x ) {
                //Bounce X
                vect.multX( bounceF );
            }
            if ( (int)oldPos.y != (int)pos.y ) {
                //Bounce Y
                vect.multY( bounceF );
            }
            if ( (float)g_game->getMap()->getPointHeight( oldPos ) >= pos.z ) {
                //Bounce Z
                vect.multZ( bounceF );
            }

            //We hit the ground, so bounce!
            pos = oldPos;

        } else {
            //Since we updated pos, we have to update our cached Rect.
            cached.x = pos.x;
            cached.y = pos.y;
        }

    } else {
        //Backup!
        pos = oldPos;
    }

    life -= dt;
    return (life <= 0.0f);
}

void Particle::draw( const Point& anchor ) {
    v[0].x = anchor.x;
    v[0].y = anchor.y;
    v[0].color = color;
    v[1].x = anchor.x + size;
    v[1].y = anchor.y;
    v[1].color = color;
    v[2].x = anchor.x;
    v[2].y = anchor.y + size;
    v[2].color = color;
    v[3].x = anchor.x + size;
    v[3].y = anchor.y + size;
    v[3].color = color;

    LPDIRECT3DDEVICE8& d = g_dx->getDeviceHandle();
    HRESULT r = 0;
    //DWORD oldShader;
    //DWORD oldBlend;
    //DWORD oldSrcBlend;
    //DWORD oldDestBlend;

    //Save old values
    //r = d->GetVertexShader( &oldShader );
    //assert( SUCCEEDED( r ) );
/*
    r = d->GetRenderState( D3DRS_SRCBLEND, &oldSrcBlend );
    assert( SUCCEEDED( r ) );

    r = d->GetRenderState( D3DRS_DESTBLEND, &oldDestBlend );
    assert( SUCCEEDED( r ) );

    r = d->GetRenderState( D3DRS_ALPHABLENDENABLE, &oldBlend );
    assert( SUCCEEDED( r ) );

    //Set our values
    r = d->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
    assert( SUCCEEDED( r ) );

    r = d->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
    assert( SUCCEEDED( r ) );

    r = d->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
    assert( SUCCEEDED( r ) );
*/
    r = d->SetVertexShader( ParticleVertexFVF );
    assert( SUCCEEDED( r ) );

    r = d->SetTexture( 0, NULL );
    assert( SUCCEEDED( r ) );

    /*
    r = d->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
    assert( SUCCEEDED( r ) );
    */

    r = d->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, v, sizeof( v[0] ) );
    assert( SUCCEEDED( r ) );

    //Restore old values

    //r = d->SetVertexShader( oldShader );
    //assert( SUCCEEDED( r ) );
/*

    r = d->SetRenderState( D3DRS_SRCBLEND, oldSrcBlend );
    assert( SUCCEEDED( r ) );

    r = d->SetRenderState( D3DRS_DESTBLEND, oldDestBlend );
    assert( SUCCEEDED( r ) );

    r = d->SetRenderState( D3DRS_ALPHABLENDENABLE, oldBlend );
    assert( SUCCEEDED( r ) );
*/}

float Particle::getHeight() const {
    return pos.z;
}

bool Particle::needsTransform() const {
    return true;
}

const Rect& Particle::getPosition() const {
    return cached;
}
