/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#include "IsoBomb.h"
#include "Powerup.h"
#include "TextureManager.h"
#include "Map.h"
#include "Game.h"
#include "RespawnPowerup.h"
#include "ConfigFile.h"
#include "Virus.h"
#include "SoundSystem.h"

float Powerup::respawnTime = 5.0f;
int Powerup::virusProb = 10;

const float RADS_PER_SEC = M_PI_F;

void Powerup::initVars( ConfigFile& f ) {
    f.getVar( "Powerup", "respawnTime", respawnTime, 5.0f );
    f.getVar( "Powerup", "virusProb", virusProb, 10 );
}

Powerup::Powerup( const Rect& initPos, float initHeight, char powType ) : 
    Entity(initPos, DIRECTION_DOWNLEFT), type(powType), 
    usedUp(false), isVirus(false) {

    cachedHeight = initHeight;

    sinCount = getrndf( M_PI_F );

    anim.setOffsets( 32, 60 );
    anim.setAlphaValue( 185 );

    // Blank props by default
    props.blastSize = 0.0f;
    props.detonationDelay = 0.0f;
    props.homingSpeed = 0.0f;
    props.hp = 0;
    props.maxBombs = 0;
    props.maxPower = 0.0f;
    props.numHoming = 0;
    props.powerRate = 0.0f;
    props.speed = 0.0f;

    switch( type ) {
    case HP_UP:
        props.hp = 1;
        anim.setAnimationFrames( g_am->getHitPointTexture() );
        break;
    case SPEED_UP:
        props.speed = 0.35f;
        anim.setAnimationFrames( g_am->getSpeedPowerUpTexture() );
        break;
    case BOMBS_UP:
        props.maxBombs = 1;
        anim.setAnimationFrames( g_am->getBombPowerUpTexture() );
        break;
    case BLAST_UP:
        props.blastSize = 1;
        anim.setAnimationFrames( g_am->getRadiusPowerUpTexture() );
        break;
    case THROW_UP:
        props.maxPower = 0.5f;
        anim.setAnimationFrames( g_am->getPowerPowerUpTexture() );
        break;
    case HOMING_UP:
        props.numHoming = 1;
        anim.setAnimationFrames( g_am->getHomingBombTexture() );
        break;
    default:
        assert( true );
    }

    // Decide if this powerup is infected with a Virus
    int r = getrnd( 100 );
    if( r <= virusProb ) {
        isVirus = true;
        TRACE2( "virus at (%f,%f)", pos.x, pos.y );
    }
}

Powerup::~Powerup() {
}

bool Powerup::update( float dt ) {
    if( usedUp ) {
        g_game->addLogic( new RespawnPowerup( new Powerup( pos, cachedHeight, type ), respawnTime ) );
        if ( !isVirus )
            g_snd->playPowerupPickupSound();
    }

    sinCount += dt;
    int yValue = (int)( 20.0f * (float)sin( RADS_PER_SEC * sinCount ) ) + 60;

    anim.setOffsets( 32, yValue );

    return usedUp;
}

void Powerup::draw( const Point& anchor ) {
    anim.draw( anchor );
}

void Powerup::affectCharacter( Character* c ) {
    if( !usedUp ) {
        Virus* v = c->getVirus();
        // Check to see if the character has bomb
        // constipation
        bool noBombs = false;
        if( v != NULL ) {
            noBombs = (v->getVirusType() == NOBOMBS_VIRUS) &&
                        (type == BOMBS_UP || type == HOMING_UP);
        }

        // Give virus if this is a virus
        if( isVirus ) {
            usedUp = (v == NULL);
            if( usedUp ) {
                // Randomly choose virus type
                int v = (int)(rand() % NUM_VIRUSES);
                c->affectMe( new Virus(v, c) );
            }
        }

        // Use powerup as normal if can't use virus
        // and it's not bomb constipation.
        if( !usedUp && !noBombs ) {
           // These are permanant powerups
            usedUp = c->affectMe( props );
        }
    }
}

float Powerup::getHeight() const {
    return cachedHeight;
}

char Powerup::getPowerupType() const {
    return type;
}
