#include "IsoBomb.h"

/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#include "SoundSystem.h"
#include "IsoException.h"

const int NUM_HOMING = 5;

SoundSystem::SoundSystem()
: manager(NULL), homingSounds( NULL ), usedHoming( NULL ), hurtCycle( 0 ) {

    manager = new CSoundManager();
    assert( g_win != NULL );
    manager->SetHwnd( g_win->getHwnd() );

    try {
        if ( manager->Initialize() != TRUE ) {
            throw IsoException("SoundSystem: Error Setting Up DirectSound");
        }

        if ( !manager->LoadWaveFile( "sfx/expl00.wav", &explosion[0] ) )
            throw IsoException( "SoundSystem: Could not load sfx/expl00.wav" );
        if ( !manager->LoadWaveFile( "sfx/expl01.wav", &explosion[1] ) )
            throw IsoException( "SoundSystem: Could not load sfx/expl01.wav" );
        if ( !manager->LoadWaveFile( "sfx/expl00.wav", &explosion[2] ) )
            throw IsoException( "SoundSystem: Could not load sfx/expl00.wav" );
        if ( !manager->LoadWaveFile( "sfx/expl01.wav", &explosion[3] ) )
            throw IsoException( "SoundSystem: Could not load sfx/expl01.wav" );

        homingSounds = new DWORD[ NUM_HOMING ];
        usedHoming = new bool[ NUM_HOMING ];
        int c;
        for ( c = 0; c < NUM_HOMING; ++c ) {
            if ( !manager->LoadWaveFile( "sfx/homing.wav", &homingSounds[c] ) )
                throw IsoException( "SoundSystem: Could not load sfx/homing.wav" );
            if ( !manager->SetPlayMode( homingSounds[c], PMODE_LOOP ) )
                throw IsoException( "SoundSystem: Could not set looping mode." );
            usedHoming[ c ] = false;
        }

        for ( c = 0; c < 2; ++c ) {
            if ( !manager->LoadWaveFile( "sfx/ow.wav", &hurtSounds[c] ) )
                throw IsoException( "SoundSystem: Could not load sfx/ow.wav" );
        }

        if ( !manager->LoadWaveFile( "sfx/death.wav", &deathSound ) )
            throw IsoException( "SoundSystem: Could not load sfx/death.wav" );

        for ( c = 0; c < SS_NUM_PSPAWN; ++c ) {
            if ( !manager->LoadWaveFile( "sfx/powerspawn.wav", &pSpawn[c] ) )
                throw IsoException( "SoundSystem: Could not load sfx/powerspawn.wav" );
        }

        for ( c = 0; c < SS_NUM_PPICKUP; ++c ) {
            if ( !manager->LoadWaveFile( "sfx/powerpickup.wav", &pPickup[c] ) )
                throw IsoException( "SoundSystem: Could not load sfx/powerpickup.wav" );
        }

        for ( c = 0; c < SS_NUM_VOICES; ++c ) {
            if ( !manager->LoadWaveFile( "sfx/bombarrhea.wav", &bombarrhea[c] ) )
                throw IsoException( "SoundSystem: Could not load sfx/bombarrhea.wav" );
            if ( !manager->LoadWaveFile( "sfx/slowdown.wav", &slowdown[c] ) )
                throw IsoException( "SoundSystem: Could not load sfx/slowdown.wav" );
            if ( !manager->LoadWaveFile( "sfx/slowbombs.wav", &slowbombs[c] ) )
                throw IsoException( "SoundSystem: Could not load sfx/slowbombs.wav" );
            if ( !manager->LoadWaveFile( "sfx/frailty.wav", &frailty[c] ) )
                throw IsoException( "SoundSystem: Could not load sfx/frailty.wav" );
            if ( !manager->LoadWaveFile( "sfx/shrink.wav", &shrink[c] ) )
                throw IsoException( "SoundSystem: Could not load sfx/shrink.wav" );
            if ( !manager->LoadWaveFile( "sfx/constipation.wav", &constipation[c] ) )
                throw IsoException( "SoundSystem: Could not load sfx/constipation.wav" );
        }

    } catch (...) {
        delete manager;
        delete[] homingSounds;
        delete[] usedHoming;
        throw;
    }
}

SoundSystem::~SoundSystem() {
    delete manager;
    delete[] homingSounds;
    delete[] usedHoming;
}

bool SoundSystem::isTokenValid( SoundToken tok ) {
    return ( (int)tok >= 0 );
}

SoundToken SoundSystem::getInvalidToken() {
    return SoundToken( -1 );
}

void SoundSystem::playExplosion() {
    assert( manager != NULL );

    BOOL r = manager->StartPlay( explosion[ rand() % 4 ] );
    assert( r == TRUE );
}

void SoundSystem::playHurtSound() {
    assert( manager != NULL );

    BOOL r = manager->StartPlay( hurtSounds[ hurtCycle % 2 ] );
    assert( r == TRUE );
    ++hurtCycle;
}

void SoundSystem::playDeathSound() {
    assert( manager != NULL );

    BOOL r = manager->StartPlay( deathSound );
    assert( r == TRUE );
}

void SoundSystem::playIfAvailable( DWORD* sounds, int numSounds ) {
    for ( int c = 0; c < numSounds; ++c ) {
        BOOL isPlaying;
        BOOL r = manager->IsPlaying( sounds[c], &isPlaying );
        assert( r == TRUE );
        if ( !isPlaying ) {
            r = manager->StartPlay( sounds[c] );
            assert( r == TRUE );
            return;
        }
    }
}

void SoundSystem::playPowerupRespawnSound() {
    playIfAvailable( pSpawn, SS_NUM_PSPAWN );
}

void SoundSystem::playPowerupPickupSound() {
    playIfAvailable( pPickup, SS_NUM_PPICKUP );
}

void SoundSystem::playBombarrheaVoice() {
    playIfAvailable( bombarrhea, SS_NUM_VOICES );
}

void SoundSystem::playFrailtyVoice() {
    playIfAvailable( frailty, SS_NUM_VOICES );
}

void SoundSystem::playShrinkVoice() {
    playIfAvailable( shrink, SS_NUM_VOICES );
}

void SoundSystem::playSlowdownVoice() {
    playIfAvailable( slowdown, SS_NUM_VOICES );
}

void SoundSystem::playSlowBombsVoice() {
    playIfAvailable( slowbombs, SS_NUM_VOICES );
}

void SoundSystem::playBombConstipationVoice() {
    playIfAvailable( constipation, SS_NUM_VOICES );
}

SoundToken SoundSystem::startHomingBombSound() {
    assert( manager != NULL );

    //Find a free homing sound slot
    int retTok = -1;
    for ( int c = 0; c < NUM_HOMING; ++c ) {
        if ( !usedHoming[ c ] ) {
            retTok = c;
            break;
        }
    }

    if ( retTok != -1 ) {
        usedHoming[ retTok ] = true;
        BOOL r = manager->StartPlay( homingSounds[ retTok ] );
        assert( r == TRUE );
    }

    return SoundToken( retTok );
}

void SoundSystem::stopHomingBombSound( SoundToken& tok ) {
    assert( manager != NULL );

    if ( tok >= 0 && tok < NUM_HOMING ) {
        BOOL r = manager->StopPlay( homingSounds[ tok ] );
        assert( r == TRUE );
        usedHoming[ tok ] = false;
    }

    tok = getInvalidToken();
}
