#include "IsoBomb.h"

/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#ifndef SOUNDSYSTEM_H_KWER234
#define SOUNDSYSTEM_H_KWER234

/**
 * The SoundSystem class returns tokens when playing looping sounds that need
 * to be stopped.  When starting a looping sound, you must return the token
 * to the cooresponding stop method when you want the sound to stop.
 *
 * A SoundToken can be either valid or invalid.  Static methods in the
 * SoundSystem class work with SoundToken objects.
 */
typedef int SoundToken;

const int SS_NUM_PSPAWN = 3;
const int SS_NUM_PPICKUP = 2;
const int SS_NUM_VOICES = 2;

/**
 * The SoundSystem is a super-high level interface for playing sounds.  It is
 * the interface between the CSoundManager library that Jason Winnebeck located
 * on codeproject.com.  This interface is not abstract, and is specific to
 * Super IsoBomb.
 */
class SoundSystem {
public:
    /**
     * Installs the sound drivers and loads all sound drivers.  Throws an
     * IsoException on error.  The sound files are loaded from the "sfx"
     * directory.
     *
     * The global g_win must be properly initialized.
     */
    SoundSystem();

    /**
     * Shuts down the SoundSystem, and uninitalizes DirectSound.
     */
    ~SoundSystem();

    /**
     * Returns true if the SoundToken is a valid token.
     */
    static bool isTokenValid( SoundToken tok );

    /**
     * Returns a SoundToken that is invalid (has an invalid value).  This can
     * be considered the token equivalent of a "NULL" pointer.
     */
    static SoundToken getInvalidToken();

    /**
     * Plays an explosion sound.
     */
    void playExplosion();

    /**
     * Plays a sound for when a player gets hurt.
     */
    void playHurtSound();

    /**
     * Plays a sound for when a player dies.
     */
    void playDeathSound();

    /**
     * Plays a sound for when a Powerup is about to respawn.
     */
    void playPowerupRespawnSound();

    /**
     * Plays a sound for when a Powerup has been picked up.
     */
    void playPowerupPickupSound();

    /**
     * Plays the voice for a bombarrhea Virus.
     */
    void playBombarrheaVoice();

    /**
     * Plays the voice for a frailty Virus.
     */
    void playFrailtyVoice();

    /**
     * Plays the voice for a shrink Virus.
     */
    void playShrinkVoice();

    /**
     * Plays the voice for a slowdown Virus.
     */
    void playSlowdownVoice();

    /**
     * Plays the voice for a slow bombs Virus.
     */
    void playSlowBombsVoice();

    /**
     * Plays the voice for a constipation Virus.
     */
    void playBombConstipationVoice();

    /**
     * Starts the homing bomb sound.  Should be started when the homing bomb
     * appears, and stopped when the homing bomb dissapears.
     *
     * If no sounds are left to be used, an invalid token is returned, and no
     * additional sound is played.  For ease of use, you need not check if
     * the returned token is valid, since the stopHomingBombSound method will
     * be able to handle the returned token that is invalid.
     */
    SoundToken startHomingBombSound();

    /**
     * Called when a homing bomb dissapears.  Changes the passed in token to
     * an invalid token.  If the passed token is invalid, this method has no
     * effect.
     *
     * @see startHomingBombSound
     */
    void stopHomingBombSound( SoundToken& tok );

private:
    /**
     * Private helper function to play sounds from the bank of available (not
     * playing) sounds.
     */
    void playIfAvailable( DWORD* sounds, int numSounds );

    CSoundManager* manager;

    DWORD explosion[4];     //explosion sounds

    DWORD* homingSounds;    //homing bomb sound loop
    bool* usedHoming;       //which sounds are in use

    DWORD hurtSounds[2];    //hurt sounds
    int hurtCycle;

    DWORD deathSound;

    DWORD pSpawn[SS_NUM_PSPAWN];//Powerup respawn sounds

    DWORD pPickup[SS_NUM_PPICKUP]; //Powerup pickup sounds

    //Virus Voices
    DWORD bombarrhea[SS_NUM_VOICES];
    DWORD slowdown[SS_NUM_VOICES];
    DWORD slowbombs[SS_NUM_VOICES];
    DWORD frailty[SS_NUM_VOICES];
    DWORD shrink[SS_NUM_VOICES];
    DWORD constipation[SS_NUM_VOICES];
};

#endif
