/**
 * Super IsoBomb
 *
 * Authors: Jim Clase, Jonathan Hilliker, Jason Winnebeck
 */

#ifndef _VIRUS_H_
#define _VIRUS_H_

#include "IsoBomb.h"
#include "CharacterProps.h"

class ConfigFile;
class Character;

// Number of viruses
const int NUM_VIRUSES =         6;
// Virus types
const int SLOW_VIRUS =          0;
const int NOBOMBS_VIRUS =       1;
const int BOMBARRHEA_VIRUS =    2;
const int WEAKTHROW_VIRUS =     3;
const int DELAY_VIRUS =         4;
const int SHRINK_VIRUS =        5;
// The maximum red value
const float MAX_RED_VALUE =     255.0f;
// The minimum red value
const float MIN_RED_VALUE =     200.0f;
// The rate at which the red value changes
const float RED_CHANGE_RATE =   150.0f;

/**
 * Affects a character in some way.  The effects only last for
 * a period of time.  Certain powerups contain viruses that 
 * are given to a Character.  Viruses can also be spread when
 * an infected player collides with the other player.  A character
 * can only have one virus at a time.
 */
class Virus {

public:

    /**
     * Default constructor
     */
    Virus();

    /**
     * Constructor
     *
     * @param virType                   type of virus
     * @param c                         pointer to the character to effect
     */
    Virus( char virType, Character* c );

    /**
     * Constructor
     *
     * @param virType                   type of virus
     * @param c                         pointer to the character to effect
     * @param timeLeft                         time left for virus to live
     */
    Virus( char virType, Character* c, float timeLeft );

    /**
     * Initialises this object
     *
     * @param virType                   type of virus
     * @param c                         pointer to the character to effect
     * @param timeLeft                  time left for virus to live
     */
    void init( char virType, Character* c, float timeLeft );

    /**
     * Destructor
     */
    virtual ~Virus();

    /**
     * Update the game logic.
     *
     * @param dt                the amount of time since the last frame
     * @returns                 true if this entity should be removed from the game
     */
    bool update( float dt );

    /**
     * Returns the value of the red channel to use when drawing the character 
     * this virus affects.
     *
     * @returns                 value of the red channel to use 
     */
    int getRedValue() const;

    /**
     * Returns a new virus that shares the same properties as this Virus
     * except it is targeted for a different character.
     *
     * @param c                 character to attach to
     * @returns                 pointer to a cloned virus
     */
    Virus* getCloneFor( Character* c ) const;

    /**
     * Returns the type of the virus.
     *
     * @returns                 type of the virus
     */
    int getVirusType() const;

    /**
     * Initializes configurable members for the Virus class.
     *
     * @param f                 a config file
     */
    static void initVars( ConfigFile& f );

private:

    /**
     * How long the virus' affect leasts
     */
    static float lifeTime;

    /**
     * The speed to reduce players to when they have the 
     * slow virus.
     */
    static float slowSpeed;

    /**
     * How long the detonation delay is for bombs for the delay virus.
     */
    static float delayTime;

    /**
     * The maximum throw power for the weak virus.
     */
    static float weakPower;

private:


    /**
     * How long until this virus expires
     */
    float countdown;

    /**
     * The props to apply to the character when first infected
     */
    CharacterProps props;

    /**
     * The character to affect
     */
    Character* host;

    /**
     * The type of powerup that contianed the virus
     */
    char type;

    /**
     * How red to draw the character who the virus inflicts
     */
    float red;

    /**
     * Flag to indicate if color should be increasing
     */
    bool increaseColor;

    /**
     * Time for next scheduled bomb throw (for bombarrhea)
     */
    float throwTime;

};

#endif